#pragma ident "$Id: recv.c,v 1.9 2004/06/04 22:49:39 dechavez Exp $"
/*======================================================================
 * 
 * Receive a frame from the peer.
 *
 *====================================================================*/
#include "iacp.h"
#include "util.h"

/* Set various values in the handle's recv field */

static VOID SetRecvError(IACP *iacp, int value)
{
    MUTEX_LOCK(&iacp->mutex);
        iacp->recv.error = value;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID SetRecvStatus(IACP *iacp, int value)
{
    MUTEX_LOCK(&iacp->mutex);
        iacp->recv.status = value;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID UpdateRecvTstamp(IACP *iacp)
{
time_t now;

    now = time(NULL);
    MUTEX_LOCK(&iacp->mutex);
        iacp->recv.tstamp = now;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID IncrRecvNbyte(IACP *iacp, int increment)
{
    MUTEX_LOCK(&iacp->mutex);
        iacp->recv.nbyte += increment;
    MUTEX_UNLOCK(&iacp->mutex);
}

static VOID IncrRecvNframe(IACP *iacp)
{
    MUTEX_LOCK(&iacp->mutex);
        ++iacp->recv.nframe;
    MUTEX_UNLOCK(&iacp->mutex);
}

/* Some errors should be ignored */

static BOOL IgnorableError(int error)
{
    switch (error) {
      case EWOULDBLOCK:
#if (EWOULDBLOCK != EAGAIN)
      case EAGAIN:
#endif
      case EINTR:
        return TRUE;
    }

    return FALSE;
}

/* Attempt to read some bytes */

static size_t ReadBytes(IACP *iacp, int LogLevel, void *ptr, size_t count)
{
size_t ReadResult;

#ifdef  _WIN32
    ReadResult = recv(iacp->sd, ptr, count,0);
#else
    ReadResult = read(iacp->sd, ptr, count);
#endif

    if (ReadResult == 0) errno = ECONNRESET;
    SetRecvError(iacp, errno);

    if (ReadResult > 0) return ReadResult;

    SetRecvStatus(iacp, iacpErrorLevel(errno));
    logioMsg(iacp->lp, LogLevel, "%s: read: %s", iacpPeerIdent(iacp), strerror(errno));

    return 0;
}


/* Local function for doing timeout enabled reads.  This assumes that
 * the file descripter has already been set for non-blocking I/O,
 * which is done as part of iacpOpen() and iacpAccept().
 */

static BOOL read_to(IACP *iacp, UINT8 *buf, UINT32 want)
{
size_t remain;
ssize_t ReadResult;
UINT8 *ptr;
int SelectResult;
int LogLevel;
fd_set readfds;
UINT32 timeoutUS;
#define USEC_PER_MSEC 1000
#define USEC_PER_SEC  1000000
struct timeval timeout;
static char *fid = "read_to";

    LogLevel = iacpGetDebug(iacp) ? LOG_ERR : LOG_DEBUG;

    if (want == 0) return TRUE;

    SelectResult = 0;
    timeoutUS = iacpGetTimeoutInterval(iacp) * USEC_PER_MSEC;
    timeout.tv_sec  = timeoutUS / USEC_PER_SEC;
    timeout.tv_usec = timeoutUS - (timeout.tv_sec * USEC_PER_SEC);

/* Mask out our file descriptor as the only one to look at */

    FD_ZERO(&readfds);
    FD_SET(iacp->sd, &readfds);
    
/*  Read from socket until desired number of bytes acquired  */

    remain = want;
    ptr    = buf;

    while (remain) {
        SelectResult = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout);
        SetRecvError(iacp, errno);
        if (SelectResult > 0) {
            ReadResult = ReadBytes(iacp, LogLevel, (void *) ptr, (size_t) remain);
            if (ReadResult > 0) {
                remain -= ReadResult;
                ptr    += ReadResult;
            } else if (!IgnorableError(errno)) {
                logioMsg(iacp->lp, LogLevel, "%s: %s: ReadBytes: %s", iacpPeerIdent(iacp), fid, strerror(errno));
                return FALSE;
            } else {
                logioMsg(iacp->lp, LOG_ERR, "%s: %s: ReadBytes: %s (IGNORED)", iacpPeerIdent(iacp), fid, strerror(errno));
                return FALSE;
            }
        } else if (SelectResult == 0) {
            errno = ETIMEDOUT;
            SetRecvError(iacp, errno);
            SetRecvStatus(iacp, IACP_ERR_TRANSIENT);
            return FALSE;
        } else if (!IgnorableError(errno)) {
            logioMsg(iacp->lp, LogLevel, "%s: %s: select: %s", iacpPeerIdent(iacp), fid, strerror(errno));
            SetRecvStatus(iacp, iacpErrorLevel(errno));
            return FALSE;
        } else {
            logioMsg(iacp->lp, LOG_ERR, "%s: %s: select: %s (IGNORED)", iacpPeerIdent(iacp), fid, strerror(errno));
        }
    }

    UpdateRecvTstamp(iacp);
    SetRecvError(iacp, 0);
    SetRecvStatus(iacp, IACP_ERR_NONE);
    IncrRecvNbyte(iacp, want);
    return TRUE;
}

/* Read and unpack a UINT32 */

static BOOL ReadUINT32(IACP *iacp, UINT32 *dest)
{
UINT32 ltmp;

    if (iacp == NULL || dest == NULL) {
        errno = EINVAL;
        return FALSE;
    }

    if (!read_to(iacp, (UINT8 *) &ltmp, 4)) return FALSE;
    utilUnpackUINT32((UINT8 *) &ltmp, dest);

    return TRUE;
}

/* Receive an IACP frame */

BOOL iacpRecvFrame(IACP *iacp, IACP_FRAME *frame, UINT8 *buf, UINT32 buflen)
{
UINT32 ltmp;
static char *fid = "iacpRecvFrame";

    if (iacp == (IACP *) NULL || frame == (IACP_FRAME *) NULL) {
        errno = EINVAL;
        return FALSE;
    }

/* Read the IACP leader */

    if (!read_to(iacp, (UINT8 *) &ltmp, 4)) return FALSE;
    if (memcmp(&ltmp, IACP_TAG, IACP_TAGLEN) != 0) {
        errno = EPROTO;
        return FALSE;
    }

/* Read the sequence number */

    if (!ReadUINT32(iacp, &frame->seqno)) return FALSE;

/* Read the payload type */

    if (!ReadUINT32(iacp, &frame->payload.type)) return FALSE;

/* Read the payload length */

    if (!ReadUINT32(iacp, &frame->payload.len)) return FALSE;

/* Read the payload */

    if (frame->payload.len > buflen) {
        logioMsg(iacp->lp, LOG_ERR, "%s: %s: payload too big (len=%lu, buflen=%lu)!",
            iacpPeerIdent(iacp), fid,  frame->payload.len, buflen
        );
        errno = EBADMSG;
        SetRecvError(iacp, errno);
        SetRecvStatus(iacp, IACP_ERR_FATAL);
        return FALSE;
    } else {
        frame->payload.data = buf;
        if (!read_to(iacp, frame->payload.data, frame->payload.len)) return FALSE;
    }
    if (iacpGetDebug(iacp) > 1) {
        logioMsg(iacp->lp, LOG_INFO, "%s: %s: frame->seqno=%lu, frame->payload.type=%lu, frame->payload.len=%lu",
            iacpPeerIdent(iacp), fid, frame->seqno, frame->payload.type, frame->payload.len
        );
    }
    if (iacpGetDebug(iacp) > 2) {
        logioHexDump(iacp->lp, LOG_INFO, frame->payload.data, frame->payload.len);
    }

/* Read the signature */

    if (!ReadUINT32(iacp, &frame->auth.id)) return FALSE;
    if (!ReadUINT32(iacp, &frame->auth.len)) return FALSE;
    if (frame->auth.len > IACP_MAXSIGLEN) {
        logioMsg(iacp->lp, LOG_ERR, "%s: %s: signature block too big (len=%lu)!",
            iacpPeerIdent(iacp), fid,  frame->auth.len
        );
        errno = EBADMSG;
        SetRecvError(iacp, errno);
        SetRecvStatus(iacp, IACP_ERR_FATAL);
        return FALSE;
    } else {
        if (!read_to(iacp, frame->auth.sig, frame->auth.len)) return FALSE;
    }
    if (iacpGetDebug(iacp) > 1) {
        logioMsg(iacp->lp, LOG_INFO, "%s: %s: frame->seqno=%lu, frame->auth.id=%lu, frame->auth.len=%lu",
            iacpPeerIdent(iacp), fid, frame->seqno, frame->auth.id, frame->auth.len
        );
    }
    if (iacpGetDebug(iacp) > 2) {
        logioHexDump(iacp->lp, LOG_INFO, frame->auth.sig, frame->auth.len);
    }

/* Verify the data */

    iacpVerifyFrame(frame);

/* Increment the packet count and we are done */

    IncrRecvNframe(iacp);

    return TRUE;
}

/* Revision History
 *
 * $Log: recv.c,v $
 * Revision 1.9  2004/06/04 22:49:39  dechavez
 * various AAP windows portability modifications
 *
 * Revision 1.8  2004/01/29 18:45:50  dechavez
 * cleaned up logic in read/write loops, log ignored errors
 *
 * Revision 1.7  2003/12/04 23:23:57  dechavez
 * added EWOULDBLOCK to list of ignored select failures, fixed bug causing
 * unneeded failures for ignored error codes
 *
 * Revision 1.6  2003/11/19 23:38:58  dechavez
 * tuned debug levels
 *
 * Revision 1.5  2003/11/04 20:03:15  dechavez
 * adjust timeout setup to account for switch from sec to msec
 *
 * Revision 1.4  2003/11/04 00:32:06  dechavez
 * tuned logging verbosity
 *
 * Revision 1.3  2003/10/16 16:36:31  dechavez
 * Many bug fixes and enhancements, to numerous to mention.
 *
 * Revision 1.2  2003/06/30 18:47:24  dechavez
 * rework retry loop to allow for changes in retry attribute to be respected,
 * rework error handling logic and log error messages instead of error codes
 *
 * Revision 1.1  2003/06/09 23:50:27  dechavez
 * initial release
 *
 */
