#pragma ident "$Id: list.c,v 1.5 2004/06/24 17:07:24 dechavez Exp $"
/*======================================================================
 *
 *  General linked list management.  
 *
 *  Includes pieces lifted essentially intact from Robert Banfill's
 *  linked list library (thanks!).
 *
 *  MT unsafe!
 *
 *====================================================================*/
#include "util.h"
#include "list.h"

VOID listPrintNode(LIST_NODE *node)
{
    printf("node=0x%08x:",    node);
    printf(" prev=0x%08x",    node->prev);
    printf(" next=0x%08x",    node->next);
    printf(" length=%lu",     node->length);
    printf(" payload=0x%08x", node->payload);
    printf(" parent=0x%08x",  node->payload);
    printf("\n");
}

VOID listPrintList(LIST *list)
{
LIST_NODE *crnt;

    crnt = listFirstNode(list);
    while (crnt != NULL) {
        listPrintNode(crnt);
        crnt = listNextNode(crnt);
    }
}

static VOID DestroyArrayView(LIST *list)
{
    if (list->array != NULL) {
        free(list->array);
        list->array = NULL;
    }
}

static VOID IncrementNodeCount(LIST *list)
{
    ++list->count;
    DestroyArrayView(list);
}

static VOID DecrementNodeCount(LIST *list)
{
    --list->count;
    DestroyArrayView(list);
}

static LIST_NODE *DestroyNode(LIST_NODE *crnt)
{
LIST *parent;
LIST_NODE *next;

    if (crnt == NULL) return NULL;
    if (crnt->next == NULL) return NULL;
    if (crnt->prev == NULL) return NULL;

    parent = (LIST *) crnt->parent;

    /* save pointer to next node */

    next = listNextNode(crnt); /* returns NULL if no more */

    /* remove the node from the list */

    crnt->prev->next = crnt->next;
    crnt->next->prev = crnt->prev;

    /* delete it */

    if (crnt->payload != NULL) free(crnt->payload);
    free(crnt);
    DecrementNodeCount(parent);

    /* return pointer to next node */

    return next;
}

static VOID *NewNode(VOID *parent, VOID *payload, UINT32 length)
{
LIST_NODE *new;

    new = (LIST_NODE *) malloc(sizeof(LIST_NODE));
    if (new == NULL) return NULL;

    new->prev = NULL;
    new->next = NULL;
    new->length = length;
    if (new->length > 0) {
        new->payload = malloc(new->length);
        if (new->payload == NULL) return NULL;
        memcpy(new->payload, payload, length);
    } else {
        new->payload = NULL;
    }
    new->parent = parent;
    IncrementNodeCount((LIST *) new->parent);

    return new;
}

BOOL listInit(LIST *list)
{

    if (list == NULL) {
        errno = EINVAL;
        return FALSE;
    }

    list->head.prev = &list->head;
    list->head.next = &list->tail;
    list->head.length = 0;
    list->head.payload = NULL;
    list->head.parent = (VOID *) list;

    list->tail.prev = &list->head;
    list->tail.next = &list->tail;
    list->tail.length = 0;
    list->tail.payload = NULL;
    list->tail.parent = (VOID *) list;

    list->count = 0;
    list->array = NULL;

    list->FreeMe = FALSE;

    return TRUE;
}

LIST *listCreate()
{
LIST *list;

    if ((list = (LIST *) malloc(sizeof(LIST))) == NULL) return NULL;

    if (!listInit(list)) {
        free(list);
        return NULL;
    }

    list->FreeMe = TRUE;

    return list;
}

LIST_NODE *listFirstNode(LIST *list)
{
LIST_NODE *retval;

    if (list == NULL) return NULL;

    if (list->head.next == &list->tail) {
        retval = NULL; /* empty list */
    } else {
        retval = list->head.next;
    }

    return retval;
}

LIST_NODE *listLastNode(LIST *list)
{
LIST_NODE *retval;

    if (list == NULL) return NULL;

    if (list->head.next == &list->tail) {
        retval = NULL; /* empty list */
    } else {
        retval = list->tail.prev;
    }

    return retval;
}

LIST_NODE *listNextNode(LIST_NODE *crnt)
{
LIST_NODE *retval;

    if (crnt == NULL) return NULL;

    if (crnt->next->next == crnt->next) {
        retval = NULL; /* end of list */
    } else {
        crnt = crnt->next;
        retval = crnt;
    }

    return retval;
}

LIST_NODE *listPrevNode(LIST_NODE *crnt)
{
LIST_NODE *retval;

    if (crnt == NULL) return NULL;

    if (crnt->prev->prev == crnt->prev) {
        retval = NULL; /* begining of list */
    } else {
        retval = crnt->prev;
    }

    return retval;
}

LIST_NODE *listInsertAfter(LIST_NODE *crnt, VOID *payload, UINT32 length)
{
LIST_NODE *new, *next;

    if (crnt == NULL || payload == NULL) return NULL;

    if ((new = NewNode(crnt->parent, payload, length)) == NULL) return NULL;

    next = crnt->next;

    new->prev = crnt;
    new->next = crnt->next;

    crnt->next = new;
    next->prev = new;

    return new;
}

LIST_NODE *listInsertBefore(LIST_NODE *crnt, VOID *payload, UINT32 length)
{
LIST_NODE *new, *prev;

    if (crnt == NULL || payload == NULL) return NULL;

    if ((new = NewNode(crnt->parent, payload, length)) == NULL) return NULL;

    prev = crnt->prev;

    new->prev = crnt->prev;
    new->next = crnt;

    crnt->prev = new;
    prev->next = new;

    return new;
}

BOOL listAppend(LIST *list, VOID *payload, UINT32 length)
{
LIST_NODE *crnt;

    if (length == 0) return TRUE; /* ignore empty payloads */

    crnt = &list->head;
    while (crnt->next != &list->tail) crnt = crnt->next;
    return (listInsertAfter(crnt, payload, length) == NULL) ? FALSE : TRUE;
}

VOID listDestroy(LIST *list)
{
LIST_NODE *crnt;

    if (list == NULL) return;
    if ((crnt = listFirstNode(list)) == NULL) return;
    while ((crnt = DestroyNode(crnt)) != NULL);

    if (list->array != NULL) free(list->array);
    list->array = NULL;
    list->count = 0;

    if (list->FreeMe == TRUE) free(list);
}

BOOL listSetArrayView(LIST *list)
{
UINT32 count;
LIST_NODE *crnt;

    if (list == NULL) {
        errno = EINVAL;
        return FALSE;
    }

    if (list->count == 0) return TRUE;

    if (list->array != NULL) return TRUE;

    list->array = (VOID **) malloc(list->count * sizeof(VOID *));
    if (list->array == NULL) return FALSE;

    count = 0;
    crnt = listFirstNode(list);
    while (crnt != NULL) {
        if (count == list->count) { /* should never occur! */
            errno = EINVAL;
            free(list->array);
            return FALSE;
        }
        list->array[count++] = crnt->payload;
        crnt = listNextNode(crnt);
    }
    if (count != list->count) { /* should never occur! */
        errno = EINVAL;
        free(list->array);
        return FALSE;
    }

    return TRUE;
}

VOID listDump(FILE *fp, LIST *list)
{
UINT32 count = 0;
LIST_NODE *crnt;

    crnt = listFirstNode(list);
    while (crnt != NULL) {
        fprintf(fp, "%lu: len=%lu\n", count++, crnt->length);
        utilPrintHexDump(fp, crnt->payload, crnt->length);
        crnt = listNextNode(crnt);
    }
}

#ifdef DEBUG_TEST

void WalkList(LIST *list)
{
UINT32 count = 0;
LIST_NODE *crnt;

    printf("WalkList, count = %lu\n", list->count);

    crnt = listFirstNode(list);
    while (crnt != NULL) {
        printf("%lu: '%s' len=%lu\n", count++, crnt->payload, crnt->length);
        crnt = listNextNode(crnt);
    }
}

void PrintArray(LIST *list)
{
UINT32 i;
char *payload;

    if (!listSetArrayView(list)) {
        perror("listSetArrayView");
        exit(1);
    }

    printf("PrintArray, count = %lu\n", list->count);

    for (i = 0; i < list->count; i++) {
        payload = list->array[i];
        printf("%lu: '%s' len=%lu\n", i, payload, strlen(payload));
    }
}

static VOID TestDestroy(LIST_NODE *crnt)
{
LIST *list;

    list = crnt->parent;
    printf("Destroy node with payload '%s'\n", crnt->payload);
    DestroyNode(crnt);
    WalkList(list);
    PrintArray(list);
}

int main(int argc, char **argv)
{
int i, count = 0;
LIST list;
LIST_NODE *crnt;
static char string[MAXPATHLEN];

    if (argc > 1) {
        count = atoi(argv[1]);
    } else {
        fprintf(stderr, "usage: %s count\n", argv[0]);
        exit(1);
    }

    if (count < 1) {
        fprintf(stderr, "%s: bad count\n", argv[0]);
        exit(1);
    }

    listInit(&list);
    for (i = 0; i < count; i++) {
        sprintf(string, "this is element %d", i);
        if (!listAppend(&list, string, strlen(string))) {
            perror("listAppend");
            exit(1);
        }
    }

    WalkList(&list);
    PrintArray(&list);

    crnt = listFirstNode(&list);
    if (crnt == NULL) {
        fprintf(stderr, "unexpected NULL return from listFirstNode\n");
        exit(1);
    }
    crnt = listNextNode(crnt);
    TestDestroy(crnt);
    crnt = listNextNode(crnt);
    crnt = listNextNode(crnt);
    TestDestroy(crnt);

    printf("destroy entire list\n");
    listDestroy(&list);
    WalkList(&list);
    PrintArray(&list);
}

#endif /* DEBUG_TEST */

/* Revision History
 *
 * $Log: list.c,v $
 * Revision 1.5  2004/06/24 17:07:24  dechavez
 * removed unused variables (aap)
 *
 * Revision 1.4  2004/06/04 22:49:46  dechavez
 * various AAP windows portability modifications
 *
 * Revision 1.3  2003/11/19 23:30:38  dechavez
 * included util.h to calm certain compilers
 *
 * Revision 1.2  2003/11/04 00:37:19  dechavez
 * improved error checking in listCreate() and listSetArrayView()
 *
 * Revision 1.1  2003/10/16 16:06:55  dechavez
 * initial release (imported from ex-liblist)
 *
 */
