/***************************************************************************** 
 * proto_sadc.cc
 * 
 * Protocol implementation for the SADC18/20/30
 *
 * (c) 2006 Henrik Thoms - GFZ Potsdam - thoms(at)gfz-potsdam.de
 *
 * Created: 12/01/06
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. For more information, see http://www.gnu.org/
 *****************************************************************************/

#include <iomanip>
#include <vector>
#include <algorithm>
#include <cstdlib>
#include <cstddef>
#include <string>
#include <ctime>

// tmp
#include <fstream>
//

#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <error.h>

#include "qtime.h"

#include "utils.h"
#include "cppstreams.h"
#include "serial_plugin.h"
#include "plugin_channel.h"
#include "diag.h"

#include "little-endian.h"

#define PACKED __attribute__ ((packed))

using namespace std;
using namespace Utilities;
using namespace CPPStreams;
using namespace SeedlinkPlugin;

namespace {

    const char *VERSION = "09.0.0";
    const int NCHAN = 16;
    const unsigned char DATA_HEADER[]  = {
	0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
	0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f, 0x90, 0x91 };
    
/* **************************************************************************
 * Data Structures
 * **************************************************************************/
    
    // Header of the data segments
    struct SegmentHeader
    {
	unsigned char commandId;
    } PACKED;

    // Time Segment containing date and time
    struct TimeSegmentFull
    {
	unsigned char year, month, day, sec, min, hour, extra, end;
    } PACKED;

    // Time Segment containing just the date
    struct TimeSegmentPartial
    {
	unsigned char sec, min, hour, extra, end;
    } PACKED;

    // Data segment for 24 bit digitizer
    struct DataSegment24Bit
    {
	unsigned char low, middle, high, end;
    } PACKED;

    // Data segment for 18 and 16 bit digitizer
    struct DataSegment18and16Bit
    {
	unsigned char low, high, end;
    } PACKED;


/* ***************************************************************************
 * SADCdataProtocol
 * ***************************************************************************/
    
    class SADCdataProtocol: public Proto
    {
    public:
	const string _myName;
	
	SADCdataProtocol(const string &myNameInit):
	    _myName(myNameInit), _countdownInitVal(360), _countdown(_countdownInitVal),
	    _samplingRate(100), _currentNrOfChannels(0), _data(NULL),
	    _timeOut(10), _maxSamplingRate(200), _timeSegment(NULL),  _dataSegment(NULL),
	    _segmentHeaderSize(1), _timeSegmentSize(0), _dataSegmentSize(0),
	    _sadcChannels(NCHAN), decodeData(NULL) ,decodeTime(NULL)
	    {
		memset(&_segmentId, 0, _segmentHeaderSize);
	    }
	
	~SADCdataProtocol()
	    {
		if (_timeSegment != NULL) free(_timeSegment);
		if (_dataSegment != NULL) free(_dataSegment);
		if (_data != NULL) free(_data);
	    }
	
	void attach_output_channel(const string &source_id,
				   const string &channel_name, const string &station_name);
	void flush_channels();
	
	void start();
	
    private:
	const int                      _countdownInitVal;
	int                            _countdown;
	string                         _digitizerVersion;
	int                            _samplingRate;
	int                            _currentNrOfChannels;
	int                            *_data;
	int                            _timeOut;
	int                            _maxSamplingRate;
	int                            _fd;
	SegmentHeader                  _segmentId;
	void                           *_timeSegment;
	void                           *_dataSegment;
	int                            _segmentHeaderSize;
	int                            _timeSegmentSize;
	int                            _dataSegmentSize;
	vector<rc_ptr<OutputChannel> > _sadcChannels;
	
	/** Pointer to member functions */
	int (SADCdataProtocol::*decodeData)(void *);
	string (SADCdataProtocol::*decodeTime)(void *);
	void (SADCdataProtocol::*setTimeFromGPS)(const void *);
	void (SADCdataProtocol::*statusMessage)(const void *);
		
	void   init();
	void   initSegments();
	void   do_start();
	int    decodeData24Bit(void *);
	int    decodeData18Bit(void *);
	int    decodeData16Bit(void *);
	void   statusMessageTimeFull(const void*);
	string decodeTimeFull(void *);
	void   set_time_from_pc();
	void   setTimeFromGPSFull(const void *);
	void   setTimeFromGPSPartial(const void *);
	string getFirmwareVersion();
	void   setGMT(char);
	void   setTime(char sec=0x00, char min=0x00, char hour=0x00);
	void   setDate(char year=0x00, char month=0x00, char day=0x00);
	void   setCrystalErrorCompensationTrim(char lowTrim=0xFF, char medTrim=0xFF,
					       char highTrim=0xFF, char dirTrim=0xFF);
	void   setSamplingRate3Channels(int samplingRate=100);
	void   setSamplingRate4Channels(int samplingRate=100);
	void   setSamplingRate16Channels(int smaplingRate=100,
				       unsigned char channelsLow=0x00,
				       unsigned char channelsHigh=0x00);
	bool   readReturnStatus(const unsigned char acknowledgeByte);
	bool   sanityCheck(unsigned char *dataSegment, int size) {
	    if (dataSegment[size-1] < 0xF0)
		return false;
	    return true;
	}
    };
    
    /**************************************************************************
     * public definition
     **************************************************************************/
    void SADCdataProtocol::attach_output_channel(const string &source_id,
						 const string &channel_name,
						 const string &station_name)
    {
	int n;
	char *tail;

	n = strtoul(source_id.c_str(), &tail, 10);
    
	if(*tail || n >= NCHAN)
	    throw PluginADInvalid(source_id, channel_name);

	if(_sadcChannels[n] != NULL)
	    throw PluginADInUse(source_id, _sadcChannels[n]->channel_name);
	
	_sadcChannels[n] = new OutputChannel(channel_name, station_name,
					     dconf.zero_sample_limit);
    }

    void SADCdataProtocol::flush_channels()
    {
	for(int n = 0; n < NCHAN; ++n)
	{
	    if(_sadcChannels[n] == NULL) continue;
	    _sadcChannels[n]->flush_streams();
	}
    }

    void SADCdataProtocol::start()
    {
	_fd = open_port(O_RDWR);
	sync_data(_fd);

	// NOTE: modified: This might be wrong !!!
	
	// if(!dconf.use_pctime_if_no_gps)
	//    logs(LOG_INFO) << "digitizer will be enabled when GPS
	//    in-lock message is received" << "\n";
	
	// END
	
	try {
	    do_start();
	}
	catch(PluginError &e) {
	    seed_log << "closing device" << "\n";
	    close(_fd);
	    throw;
	}

	seed_log << "closing device" << "\n";
	close(_fd);
    }

    /* *************************************************************************
     * private definitions
     * *************************************************************************/

    void SADCdataProtocol::initSegments()
    {
	if  ( (_data  = (int *) malloc(NCHAN * sizeof(int))) == NULL)
	    throw bad_alloc();
	memset(_data, 0, NCHAN * sizeof(int));
	
	// allocate time segment
	if ((_timeSegment = malloc(_timeSegmentSize)) == NULL)
	    throw bad_alloc();
	memset(_timeSegment, 0, _timeSegmentSize);	    
	
	// allocate data segment
	if ((_dataSegment = malloc(_dataSegmentSize)) == NULL)
	    throw bad_alloc();
	memset(_dataSegment, 0, _dataSegmentSize);
    }
    
    /**
     * Configures the plugin in correspondence to the firmaware version.
     */
    void SADCdataProtocol::init()
    {
        _digitizerVersion = getFirmwareVersion();
	logs(LOG_INFO) << "Firmwareversion: " << _digitizerVersion << "\n";

	// We are using a gps time source, so no GMT corection is
	// necessary. 
	setGMT(0);
	
	// Set default time and date.
	setTime();
	setDate(0x00, 0x01, 0x01);

	setCrystalErrorCompensationTrim();
	    
	if (_digitizerVersion == "V300")
	{
	    _currentNrOfChannels = 16;
	    
	    // setting up pointers to member functions
	    decodeData = &SADCdataProtocol::decodeData16Bit;
	    decodeTime = &SADCdataProtocol::decodeTimeFull;
	    setTimeFromGPS = &SADCdataProtocol::setTimeFromGPSFull;
	    statusMessage = &SADCdataProtocol::statusMessageTimeFull;
	    
	    // set segment sizes
	    _segmentHeaderSize = sizeof(SegmentHeader);
	    _timeSegmentSize = sizeof(TimeSegmentFull);
	    _dataSegmentSize = sizeof(DataSegment18and16Bit);

	    _samplingRate = 100;
	    setSamplingRate16Channels(_samplingRate);
	    
	    initSegments();

	    seed_log << "SADCdataProtocol version: " << VERSION << "\n"
		     << "Digitizer firmware: " << _digitizerVersion << "\n"
		     << "Number of channels: " << _currentNrOfChannels << "\n"
		     << "Samplingrate: " << _samplingRate << "\n";
	}
	else if (_digitizerVersion == "V200")
	{
	    _currentNrOfChannels = 3;
	    
	    // setting up pointers to member functions
	    decodeData = &SADCdataProtocol::decodeData24Bit;
	    decodeTime = &SADCdataProtocol::decodeTimeFull;
	    setTimeFromGPS = &SADCdataProtocol::setTimeFromGPSFull;
	    statusMessage = &SADCdataProtocol::statusMessageTimeFull;
	    	    
	    // set segment sizes
	    _segmentHeaderSize = sizeof(SegmentHeader);
	    _timeSegmentSize = sizeof(TimeSegmentFull);
	    _dataSegmentSize = sizeof(DataSegment24Bit);
	    	    
	    _samplingRate = 100;
	    setSamplingRate3Channels(_samplingRate);

	    initSegments();

	    seed_log << "SADCdataProtocol version: " << VERSION << "\n"
		     << "Digitizer firmware: " << _digitizerVersion << "\n"
		     << "Number of channels: " << _currentNrOfChannels << "\n"
		     << "Samplingrate: " << _samplingRate << "\n";
	}
	else if (_digitizerVersion == "V181")
	{
	    _currentNrOfChannels = 4;

	    // setting up pointers to member functions
	    decodeData = &SADCdataProtocol::decodeData18Bit;
	    decodeTime = &SADCdataProtocol::decodeTimeFull;
	    setTimeFromGPS = &SADCdataProtocol::setTimeFromGPSFull;
	    statusMessage = &SADCdataProtocol::statusMessageTimeFull;
	    
	    // set segment sizes
	    _segmentHeaderSize = sizeof(SegmentHeader);
	    _timeSegmentSize = sizeof(TimeSegmentFull);
	    _dataSegmentSize = sizeof(DataSegment18and16Bit);

	    _samplingRate = 100;
	    setSamplingRate4Channels(_samplingRate);

	    initSegments();

	    seed_log << "SADCdataProtocol version: " << VERSION << "\n"
		     << "Digitizer firmware: " << _digitizerVersion << "\n"
		     << "Number of channels: " << _currentNrOfChannels << "\n"
		     << "Samplingrate: " << _samplingRate << "\n";
	}
	else if (_digitizerVersion == "V180")
	{
	    seed_log << "Digitizerversion " << _digitizerVersion << "is _NOT_ supported\n";
	    exit(0);
	}
	else if (_digitizerVersion == "V162")
	{
	    _currentNrOfChannels = 4;
	    
	    // setting up pointers to member functions
	    decodeData = &SADCdataProtocol::decodeData16Bit;
	    decodeTime = &SADCdataProtocol::decodeTimeFull;
	    setTimeFromGPS = &SADCdataProtocol::setTimeFromGPSFull;
	    statusMessage = &SADCdataProtocol::statusMessageTimeFull;
	    
	    // set segment sizes
	    _segmentHeaderSize = sizeof(SegmentHeader);
	    _timeSegmentSize = sizeof(TimeSegmentFull);
	    _dataSegmentSize = sizeof(DataSegment18and16Bit);

	    _samplingRate = 100;
	    setSamplingRate4Channels(_samplingRate);

	    initSegments();

	    seed_log << "SADCdataProtocol version: " << VERSION << "\n"
		     << "Digitizer firmware: " << _digitizerVersion << "\n"
		     << "Number of channels: " << _currentNrOfChannels << "\n"
		     << "Samplingrate: " << _samplingRate << endl;
	}
	else if (_digitizerVersion == "V161")
	{
	    seed_log << "Digitizerversion " << _digitizerVersion << "is _NOT_ supported\n";
	    exit(0);
	}
	else if (_digitizerVersion == "V151")
	{
	    seed_log << "Digitizerversion " << _digitizerVersion << "is _NOT_ supported\n";
	    exit(0);
	}
    }

    inline void SADCdataProtocol::statusMessageTimeFull(const void *timeSegment)
    {
	const TimeSegmentFull * ts = static_cast<const TimeSegmentFull*>(timeSegment);

	if ( ts->sec == 0x00 && ts->min == 0x00 ) //&& ts->hour == 0x00 )
	{
	    seed_log << "[Status Message] SADCdataProtocol version: " << VERSION << "\n"
 		     << "digitizer firmware version: " << _digitizerVersion << "\n"
 		     << "sampling rate: " << _samplingRate << endl;
	}
    }
    
    bool SADCdataProtocol::readReturnStatus(const unsigned char acknowledgeByte)
    {
	int size = sizeof(acknowledgeByte);
	unsigned char returnStatus = 0x00;
	bool success = false;
 	time_t startTime = 0, currentTime = 0, *status = NULL;
 	startTime = time(status);
		
	while (true)
	{
	    currentTime = time(status);

	    // cout << "diffTime: " << difftime(currentTime, startTime) << "\n";
	    if (difftime(currentTime, startTime) >= _timeOut) break;
	    
	    if(read_port(_fd, &returnStatus, size) == 0) continue;
	    if (returnStatus != acknowledgeByte) continue;

	    success = true;
	    break;
	}
	return success;
    }
    
    string SADCdataProtocol::getFirmwareVersion()
    {
	int commandSize = 6;
	const unsigned char command[] =
	    { 0x81, 0x00, 0x00, 0x00, 0x00, 0x00 };
	char answer[4];
	
	while (true)
	{
	    write(_fd, command, commandSize);
	    
	    if (read_port(_fd, answer, 1) == 0) continue;
	    if (answer[0] != 'V') continue;
	    if (read_port(_fd, &(answer[1]), 3) == 0)
		logs(LOG_WARNING) << "Could not read firmware version\n";
	    break;
	}
	return string(answer, answer + 4);
    }

    void SADCdataProtocol::setTimeFromGPSFull(const void *timeSegmentFull)
    {
	const TimeSegmentFull *timeSegment = static_cast<const TimeSegmentFull*>(timeSegmentFull);
	EXT_TIME et;
	
	et.year = timeSegment->year + 2000;
	et.month = timeSegment->month;
	et.day = timeSegment->day;
	et.hour = timeSegment->hour;
	et.minute = timeSegment->min;
	et.second = timeSegment->sec;
	et.usec = 0;
	et.doy = mdy_to_doy(et.month, et.day, et.year);
	digitime.it = add_time(ext_to_int(et), 0, dconf.time_offset);
    }

    void SADCdataProtocol::setTimeFromGPSPartial(const void *timeSegmentPartial)
    {
	logs(LOG_INFO) << "setTimeFromGPSPartial() NOT IMPLEMENTED YET\n";
    }
    
    /**
     * Decodes data from the 24 bit digitizer (V200)
     *
     * PARAMETER:
     * ptr:	void pointer to an DataSegment24Bit Segment
     *
     * RETURN VALUE:
     * int:	Integer which hold the decoded data
     */
    int SADCdataProtocol::decodeData24Bit(void *ptr)
    {
	DataSegment24Bit* data = static_cast<DataSegment24Bit*>(ptr);
	
	data->low = data->low + ((data->end & 1)  << 7);
	data->middle = data->middle + ((data->end & 2)  << 6);
	
  	int tmp = 0;
	int value = 0;
  	tmp = ((data->high << 16) + (data->middle << 8 ) + data->low);
  	if (data->end & 4)
	    value = -8388608 + tmp;
 	else
  	    value = tmp;
	
  	return value;
    }

    /**
     * Decodes data from the 18 bit digitizer (V180/V181)
     *
     * PARAMETER:
     * ptr:	void pointer to an DataSegment18Bit Segment
     *
     * RETURN VALUE:
     * int:	Integer which hold the decoded data
     */
    int SADCdataProtocol::decodeData18Bit(void* ptr)
    {
	DataSegment18and16Bit *data = static_cast<DataSegment18and16Bit*>(ptr);

	data->low = data->low + ((data->end & 1)  << 7);
	data->high = data->high + ((data->end & 2)  << 6);
	
	int value = (data->high << 8) + data->low;
	
	if (data->end & 4) 
	    value += 65536;
	if (data->end & 8) 
	    value += -131072;
		
	return value;
    }
    
    /**
     * Decodes data from the 16 bit digitizer (V162/161)
     *
     * PARAMETER:
     * ptr:	void pointer to an DataSegment16Bit Segment
     *
     * RETURN VALUE:
     * int:	Integer which hold the decoded data
     */
    int SADCdataProtocol::decodeData16Bit(void* ptr)
    {
	DataSegment18and16Bit *data = static_cast<DataSegment18and16Bit*>(ptr);
	
	if (data->end == 253)
	    data->low += 128;
	else if (data->end == 254)
	    data->high += 128;
	else if (data->end == 255)
	{
	    data->low += 128;
	    data->high += 128;
	}
	short value = (data->high << 8) + data->low;
	
	return value;
    }
    
    /**
     * Decodes the time and date.
     *
     * PARAMETER:
     * time:	void pointer to a TimeSegmentFull Segment
     *
     * RETURN VALUE
     * string:  A string which contains both time and date.
     */
    string SADCdataProtocol::decodeTimeFull(void *time)
    {
	TimeSegmentFull* ts = static_cast<TimeSegmentFull*>(time);
	char* formatStr = "%02d:%02d:%02d %02d.%02d.20%02d\n";
	int strLen = 21;
	char timeStr[strLen];

	snprintf(timeStr, strLen, formatStr,
		 ts->hour, ts->min, ts->sec,
		 ts->day, ts->month, ts->year);

	return string(timeStr);
    }
    
    void SADCdataProtocol::setGMT(char GMT = 0x00)
    {
	unsigned char acknowledgeByte = 0xF8;
	char commandGMT[] =
	    { 0x82, GMT, 0x00, 0x00, 0x00, 0x00 };
	int commandSize = 6;
	
	if (write(_fd, commandGMT, commandSize) == -1 )
	    logs(LOG_WARNING) << "Could not write to filedescriptor - ERROR: "
			      << strerror(errno) << "\n";

	// readReturnStatus
	if (!readReturnStatus(acknowledgeByte))
	    logs(LOG_INFO) << "[ setGMT ]  Acknowledgebyte _NOT_ found \n";
	else
	    logs(LOG_INFO) << "[ setGMT ]  Acknowledgebyte _FOUND_ \n";
    }

    void SADCdataProtocol::setTime(char sec, char min, char hour)
    {
	unsigned char ackByte = 0xF8;
	char timeCmd[] =
	    { 0x83, sec, min, hour, 0x00, 0x00 };
	int cmdSize = 6;
	
	if (write(_fd, timeCmd, cmdSize) == -1)
	    logs(LOG_WARNING) << "[ setTime ] Could not write to filedescriptor - _ERROR_: "
			      << strerror(errno) << "\n";
	// readReturnStatus
	if (!readReturnStatus(ackByte))
	    logs(LOG_INFO) << "[ setTime ]  Acknowledgebyte _NOT_ found \n";
	else
	    logs(LOG_INFO) << "[ setTime ]  Acknowledgebyte _FOUND_ \n";
    }

    void SADCdataProtocol::setDate(char year, char month, char day)
    {
	unsigned char ackByte = 0xF8;
	char dateCmd[] =
	    { 0x87, year, month, day, 0x00, 0x00 };
	int cmdSize = 6;
	
	if (write(_fd, dateCmd, cmdSize) == -1 )
	    logs(LOG_WARNING) << "[ setDate ] Could not write to filedescriptor - _ERROR_: "
			      << strerror(errno) << "\n";

	// readReturnStatus
	if (!readReturnStatus(ackByte))
	    logs(LOG_INFO) << "[ setDate ]  Acknowledgebyte _NOT_ found \n";
	else
	    logs(LOG_INFO) << "[ setDate ]  Acknowledgebyte _FOUND_ \n";
    }
    
    void SADCdataProtocol::setCrystalErrorCompensationTrim(char lowTrim, char medTrim,
							   char highTrim, char dirTrim)
    {
	unsigned char ackByte = 0xF8;
	char dateCmd[] =
	    { 0x85, lowTrim, medTrim, highTrim, dirTrim, 0x00 };
	int cmdSize = 6;
	
	if (write(_fd, dateCmd, cmdSize) == -1 )
	    logs(LOG_WARNING) << "[ setCrystalErrorCompensationTrim ]"
		"Could not write to filedescriptor - _ERROR_: "
			      << strerror(errno) << "\n";

	// readReturnStatus
	if (!readReturnStatus(ackByte))
	    logs(LOG_INFO) << "[ setCrystalErrorCompensationTrim ]  Acknowledgebyte _NOT_ found \n";
	else
	    logs(LOG_INFO) << "[ setCrystalErrorCompensationTrim ]  Acknowledgebyte _FOUND_ \n";
    }
    
    void SADCdataProtocol::setSamplingRate16Channels(int samplingRate,
						     unsigned char channelsLow,
						     unsigned char channelsHigh)
    {
 	if (samplingRate > _maxSamplingRate || samplingRate < 0)
 	{
 	    logs(LOG_WARNING) << "sample rate has to be between "
			      << _maxSamplingRate << " and 0\n";
 	    return;
 	}
	
 	char spsX = 0x00;
 	if (samplingRate > 0)
 	    spsX = static_cast<char>(_maxSamplingRate / samplingRate);

	int commandSize = 6;
	char samplingCommand[] =
	    { 0x84, spsX, channelsLow, channelsHigh, 0x00, 0x00 };
	
	if (write(_fd, samplingCommand, commandSize) == -1 )
	    logs(LOG_WARNING) << "Could not write to filedescriptor (setSamplingRate)"
			      << " - ERROR: " << strerror(errno) << "\n";
    }
    
    void SADCdataProtocol::setSamplingRate4Channels(int samplingRate)
    {
 	if (samplingRate > _maxSamplingRate || samplingRate < 0)
 	{
 	    logs(LOG_WARNING) << "sample rate has to be between " << _maxSamplingRate << " and 0\n";
 	    return;
 	}
	
 	char spsX = 0x00;
 	if (samplingRate > 0)
 	    spsX = static_cast<char>(_maxSamplingRate / samplingRate);

	int commandSize = 6;
	char samplingCommand[] =
	    { 0x84, spsX, spsX, spsX, spsX, 0x00 };
	
	if (write(_fd, samplingCommand, commandSize) == -1 )
	    logs(LOG_WARNING) << "Could not write to filedescriptor (setSamplingRate)"
		 << " - ERROR: " << strerror(errno) << "\n";
    }

     void SADCdataProtocol::setSamplingRate3Channels(int samplingRate)
     {
 	if (samplingRate > _maxSamplingRate || samplingRate < 0)
 	{
 	    logs(LOG_WARNING) << "sample rate has to be between " << _maxSamplingRate << " and 0\n";
 	    return;
 	}
	
 	char spsX = 0x00;
 	if (samplingRate > 0)
 	    spsX = static_cast<char>(_maxSamplingRate / samplingRate);

	int commandSize = 6;
	char samplingCommand[] =
	    { 0x84, spsX, spsX, spsX, 0x00, 0x00 };
	
	if (write(_fd, samplingCommand, commandSize) == -1 )
	    logs(LOG_WARNING) << "Could not write to filedescriptor (setSamplingRate)"
		 << " - ERROR: " << strerror(errno) << "\n";
    }
    
    /**
     * do_start() reads the data from the digitatizer, does the
     * processing and sends the results to the SeisComp server.
     */
    void SADCdataProtocol::do_start()
    {
	// ofstream logFile("./sinusLog.txt");
	ofstream logFile("/dev/null");
   	if (!logFile.is_open())
    	{
    	    logs(LOG_WARNING) << "Opening logFile failed\n";
    	    exit(0);
    	}
	
	bool GPSlock = false;
	digitime.quality = 50;
	
	// Initialize the plugin according to the firmware version.
       	init();
	
	while (!terminate_proc)
 	{
	    // Read the segment header (command)
	    //
	    if (read_port(_fd, &_segmentId, _segmentHeaderSize) == 0)
		continue;
	    
	    // Time command
	    if (_segmentId.commandId == 0x81)
 	    {
		logs(LOG_INFO) << "Time segment found: " << hex << "0x"
			       << static_cast<int>(_segmentId.commandId) << dec << "\n";
		logFile << "Time segment found: " << hex << "0x"
			<< static_cast<int>(_segmentId.commandId) << dec << "\n";
		
		if (read_port(_fd, _timeSegment, _timeSegmentSize) == 0)
		{
		    logs(LOG_WARNING) << "read timeout for TimeSegment" << "\n";
		    logFile << "read timeout for TimeSegment" << "\n";
		    continue;
		}
		
		if (!sanityCheck((unsigned char *)_timeSegment, _timeSegmentSize))
		{
		    logFile << "sanity check for time segment failed: timeSgmentSize: "
			    << _timeSegmentSize << "\n";
		    continue;
		}

		// Set digitime - important for seed logging/sending etc.
		(this->*setTimeFromGPS)(_timeSegment);
		digitime.valid = true;
				
		logs(LOG_INFO) << "TIME: " << (this->*decodeTime)(_timeSegment) << "\n";
		logFile << "TIME: " << (this->*decodeTime)(_timeSegment);

		// Send status message every full hour
		(this->*statusMessage)(_timeSegment);
				
		// Check wether we have a GPS lock. Otherwise we
		// should do the time calculation on our own.
		//
		if (((char *)_timeSegment)[_timeSegmentSize - 2] & 32 && !GPSlock)
		{
		    logs(LOG_INFO)
			<< "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS _IS_ locked\n";
		    logFile << "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS _IS_ locked\n";
		    seed_log << "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS _IS_ locked\n";
		    
		    _countdown = _countdownInitVal;

		    digitime.quality = 100;
		    GPSlock = true;
		}

		if (_countdown == 0)
		{
		    logs(LOG_INFO)
			<< "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS is _NOT_ locked\n";
		    logFile << "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS is _NOT_ locked\n";
		    seed_log << "[ TIME ] " + (this->*decodeTime)(_timeSegment) + " GPS is _NOT_ locked\n";

		    GPSlock = false;
		    digitime.quality = 50;
		}
		if (_countdown >= 0)
		    _countdown--;
	    }
	    else
	    {
		// Calculate channels index
		int i = static_cast<int>(_segmentId.commandId - DATA_HEADER[0]);
		
		if (i >= _currentNrOfChannels || i < 0) // invalid index
		{
		    logs(LOG_WARNING) << "UNKNOWN HEADER: " << hex << "0x"
		    		      << static_cast<int>(_segmentId.commandId)
				      << " for CHANNEL: " << dec << i << "\n";
		    logFile << "UNKNOWN HEADER: " << hex << "0x"
			    << static_cast<int>(_segmentId.commandId)
				      << " for CHANNEL: " << dec << i << "\n";
		}
		else 
		{
		    if (read_port(_fd, _dataSegment, _dataSegmentSize) == 0)
		    { 
			logs(LOG_WARNING) << "read timeout for channel: " << i << "\n";
			logFile << "read timeout for channel: " << i << "\n";
			continue;
		    }
		    if (!sanityCheck((unsigned char *)_dataSegment, _dataSegmentSize))
		    {
			logFile << "sanity check for data segment failed - CHANNEL: " << i << "\n";
			continue;
		    }
		    _data[i] = (this->*decodeData)(_dataSegment);
		    
		    //if (_segmentId.commandId == 0x83)
		    //    logFile << _data[i] << "\n";
		    
		    logs(LOG_INFO) << "CHANNEL[" << i << "]: " << _data[i] << "\n";
		    logFile << "CHANNEL[" << i << "]: " << _data[i] << "\n";
		    		    
		    if (_sadcChannels[i] == NULL)
			continue;
		    _sadcChannels[i]->set_timemark(digitime.it, 0,
						   digitime.quality);
		    _sadcChannels[i]->put_sample(_data[i]);
		}
	    }
	    // cout.flush();
	}
	logFile.close();
    }
    
    RegisterProto<SADCdataProtocol> proto("sadc");
	
} // unnamed namespace
    
