#! /usr/bin/env python

# -----------------------------------------------------------------------------------
# get_latest_event.py
# 
# The Script reads the latest eventfile and generates the plugins for reading all the
# involved stations as well as for writing the events back to the
# harddisc.
# 
# (C) Henrik Thoms - GFZ-Potsdam
#
# Created: 27/02/2006 Revision: 
# Version: 0.1.0
# ------------------------------------------------------------------------------------

import os
import glob
import shutil
import sys
import stat

from Tkinter import *

# global variables for generating the plugins
# -------------------------------------------
EVENTID = ''
STATIONCODES = []
ONSETTIMES = []

HOME                 = os.getenv('HOME')
WRITE_EVT_PY_TPL     = HOME + '/seiscomp/analysis/private/write_event.py.tpl'
READ_STREAMS_TPL     = HOME + '/seiscomp/analysis/private/READ_STREAMS.SHC.tpl'
EXTERNAL_PROCESS     = HOME + '/seiscomp/analysis/private/external_processes.txt'
EXTERNAL_PROCESS_TPL = EXTERNAL_PROCESS + '.tpl'
SFDLINE              = HOME + '/seiscomp/analysis/sh/source/seed_io/sfdline'
EVT_FILES_PATH       = HOME + '/seiscomp/autoloc/events/shm_*.evt'
SHSCRATCH_PATH       = HOME + '/seiscomp/analysis/shscratch/'

# Determines how many events should be shown to be selected
MAX_EVT_ENTRIES = 30

# widget class for selecting the found events manually
# ---------------------------------------------------- 
class EvtWindow(Canvas, Frame):
    
    def __init__(self, master = None, evtList = []):

        Frame.__init__(self, master, takefocus = 1)
        
        self._evtList = evtList[:]
        
        self.master.title('event selection')
        
        retVal = os.popen2('xwininfo -root | grep "Width" | cut -d: -f2', 'r')
        xOffset = retVal[1].read().strip()
        retVal =  os.popen2('xwininfo -root | grep "Height" | cut -d: -f2', 'r')
        yOffset = retVal[1].read().strip()
        width = 200
        height = 250
        xOffset = (int(xOffset) / 2) - (width / 2)
        yOffset = (int(yOffset) / 2) - (height / 2)
        
        # print 'height ' + str(yOffset) + ' width: ' + str(yOffset) 
        self.master.wm_geometry('%dx%d+%d+%d' % (width, height, xOffset, yOffset) )
        self.master.protocol('WM_DELETE_WINDOW', self._windowClose)
        
        #Register global event handler
        self.bind_all('<Any-KeyPress>', self._evtHandler)
        self.bind_all('<Double-Button-1>', self._myExit)
        self.grid( sticky = N+S+W+E )
        
        # This is necessary to make the window resizeable
        top = self.winfo_toplevel()                
        top.rowconfigure(0, weight = 1)
        top.columnconfigure( 0, weight = 1 )        
        
        self.rowconfigure( 0, weight = 1 )
        self.columnconfigure( 0, weight = 1 )       
        
        self.createListbox( evtList )
        self.createButtons()
    
        #self._selectedEvent = self._listbox.get(0)
        self._selectedEvent = self._evtList[0]
    
    
    def createListbox(self, evtList):
    
        # Define horizontal and vertical scrollbars
        scrollY = Scrollbar( self, orient = VERTICAL )
        scrollY.grid( row = 0, column = 2, sticky = N+S )
        scrollX = Scrollbar( self, orient = HORIZONTAL )
        scrollX.grid ( row = 1, column = 0, columnspan = 3, sticky = E+W )
        # Definition of a scrollable listbox
        self._listbox = Listbox( self, selectmode = SINGLE, 
                                xscrollcommand = scrollX.set, 
                                yscrollcommand = scrollY.set )
        
               
        for item in  map(self._evtToReadableForm, evtList):
            self._listbox.insert(END, item)
            self._listbox.grid(row = 0, columnspan = 2, sticky = N+W+E+S)

        self._listbox.selection_set( 0 )
        
        scrollX["command"]  =  self._listbox.xview
        scrollY["command"]  =  self._listbox.yview 


    def createButtons(self):

        self._okButton = Button(self, text='ok')
        self._okButton.grid(row = 2, column = 0, sticky=E+W)
        self._okButton.bind('<1>', self._myExit)
        self._okButton.bind_all('<KeyRelease-Return>', self._myExit)

        cancelButton = Button(self, text = 'cancel', command = self._windowClose)
        cancelButton.grid(row = 2, column = 1, columnspan = 2, sticky=E+W)

    
    def _evtToReadableForm(self, evtId):
        
        # eventid format example 1060315246
        evtId = evtId.split('_')[-1].split('.')[0]
        
        index = 0
        year = 1900
         
        if int( evtId[0] ) == 1:
            year += int( ''.join(evtId[index:3]))
            index = 3
        else:
            year += int( ''.join(evtId[index:2]))
            index = 2
        
        month =  ''.join(evtId[index:index+2])
        index += 2
        day =  ''.join(evtId[index:index+2])
        index += 3
        evtNum = ''.join(evtId[index:index+3])
        
        return '%d-%s-%s evt: %s' % (year, month, day, evtNum)
   
   
    # Default event handler which is registered to the root panel
    def _evtHandler(self, evt = None):
        
        if evt.keysym == 'Up' or evt.keysym == 'Down':
            listSize = self._listbox.size()
            index = int( self._listbox.curselection()[0] )
            self._listbox.selection_clear( index )

            if evt.keysym == 'Up':
                index = (index - 1) % listSize
            
            else:
                index = (index + 1) % listSize
            self._listbox.selection_set( index )
            self._listbox.see(index)
        
        elif evt.keysym == 'Return' or evt.num == 1:
            index = int(self._listbox.curselection()[0])
            #self._selectedEvent = self._listbox.get(index)
            self._selectedEvent = self._evtList[index]
            print 'Current selection is: ' + self._selectedEvent
            if evt.keysym == 'Return':
              self._okButton.config(relief = SUNKEN)
              self._okButton.update()
       
        elif evt.keysym == 'Prior' or evt.keysym == 'Next':
            listSize = self._listbox.size()
            index = int( self._listbox.curselection()[0] )
            self._listbox.selection_clear( index )

            if evt.keysym == 'Prior':
                index = (index - 5) % listSize
            else:
                index = (index + 5) % listSize
  
            self._listbox.selection_set( index )
            self._listbox.see(index)

        elif evt.keysym == 'Home':
            self._listbox.see(0)
            self._listbox.selection_set(0)
        
        elif evt.keysym == 'End':
            self._listbox.see(END)
            self._listbox.selection_set(END)
     

        elif evt.keysym == 'Escape':
            self._windowClose()
        
        else:
           print 'Event: ' + evt.keysym + ' is not registered'

    
    def _myExit(self, evt = None):

        if evt.keysym == 'Return':
            self._okButton.config(relief = RAISED)
            self._okButton.update()
            self.quit()
        elif evt.num == 1:
            self._evtHandler(evt)
            self.quit()

    def _windowClose(self):
        sfdfile = os.environ['MYDATA'] + 'sfdfile.sfd'
        if os.path.isfile(sfdfile):
           sys.exit()
        else:
            self.quit()
    
    def getSelectedEvent(self):
        return self._selectedEvent


# Offset is in seconds
# NOTE: Offset cannot be more than 60 seconds !!!
# -----------------------------------------------
TIME_OFFSET = 30

def findLatestEvtFile():

    def cmp(file0, file1):
        evt0 = os.path.basename(file0).replace('shm_', '').replace('.evt', '')
        evt1 = os.path.basename(file1).replace('shm_', '').replace('.evt', '')
        intEvt0 = int(evt0)
        intEvt1 = int(evt1)

        if intEvt0 < intEvt1: return -1
        elif intEvt0 == intEvt1: return 0
        else: return 1
             
    fileList = glob.glob( EVT_FILES_PATH )
    # print 'unsorted: '
    # for i in fileList: print i
    
    if len( fileList ) == 0:
        print """
        ========================================================
        |                 NO EVENT FILES FOUND                 |
        ========================================================
        """
        sys.exit(0)
    
    fileList.sort(cmp)
    
    fileList.reverse()
    
    # print 'sorted: '
    # for i in fileList: print i
    
    print 'The latest event is: ' + fileList[0]
    
    # Chose a event from the gui
    slice = len(fileList)
    if slice > MAX_EVT_ENTRIES:  
        slice = MAX_EVT_ENTRIES
    
    #app = EvtWindow(Tk(), map(os.path.basename, fileList[:slice]) )
    app = EvtWindow(Tk(), fileList[:slice] )
    app.mainloop()
   
    #return os.path.dirname(fileList[0]) + '/' + app.getSelectedEvent()
    return app.getSelectedEvent()

def parseEvtFile(evtFile):
    global EVENTID
    for line in evtFile:
        # do not remove the white space after the colon !
        tokens = line.split(': ')
        if len(tokens) != 2:
            continue
        key, value = tokens
        key = key.strip()
        value = value.strip()
        value = value.rstrip('\n')
        
        if key == 'Event ID':
            EVENTID = value
        elif key == 'Station code':
            STATIONCODES.append(value)
        elif key == 'Onset time':
            ONSETTIMES.append(value)

def calcTimeWindow():

    # Time format example: 08-MAR-2006_14:51:58.200
    t0 = ONSETTIMES[0].split('_')[-1].split(':')
    t1 = ONSETTIMES[-1].split('_')[-1].split(':')
    
    h0   = int( t0[0] )
    min0 = int( t0[1] )
    sec0   = int( t0[2].split('.')[0] )
    h1   = int( t1[0] )
    min1 = int( t1[1] )
    sec1   = int( t1[2].split('.')[0] )
    print 'h0: ' + str(h1) + ' min0: ' + str(min0) + ' sec0: ' + str(sec0) + \
          ' h1: ' + str(h1) + ' min1: ' + str(min1)    + ' sec1: ' + str(sec1)
    
    dh   = h1 - h0
    dmin = min1 - min0
    dsec = sec1 - sec0
    if dh < 0:
        dh += 24
    if dmin < 0:
        dmin += 60
    if dsec < 0:
        dsec += 60
    
    tw = (dh * 3600) + (dmin * 60) + dsec + (2 * TIME_OFFSET)
    
    print 'Original time window: ' + str(tw - 2*TIME_OFFSET)    
    print 'New time window: ' + str( tw )

    return tw
    

def addTimeOffset( timeStr ):

    # Time format example: 08-MAR-2006_14:51:58.200
    time = timeStr.split('_')
    tokens = time[-1].split(':')
    h = int( tokens[0] )
    min = int( tokens[1] )
    tail = tokens[-1].split('.')
    sec = int( tail[0] )
    
    sec -= TIME_OFFSET 
    if sec < 0:
        sec += 60
        if min > 0:
            min -= 1
        elif min == 0:
            if h > 0:
                h -= 1
                min = 59
            else:
                sec = 0
    
    ts = time[0] + '_' + str(h) + ':' + str(min) + ':' + str(sec) + '.' + tail[-1]
    print 'Original time string: ' + timeStr 
    print 'New time string: ' + ts

    return ts
        

def buildSfdWildcard(evtId):

    # eventid format example 1060315246
    index = 0
    year = 1900
    if int( evtId[0] ) == 1:
        year += int( ''.join(evtId[index:3]))
        index = 3
    else:
        year += int( ''.join(evtId[index:2]))
        index = 2
    month =  ''.join(evtId[index:index+2])
    index += 2
    day =  ''.join(evtId[index:index+2])
    
    print 'date for event: %s is %s-%s-%s' % (evtId, str(year), month, day) 
    return '*`date +\'%%Y.%%j\' -d \'%s-%s-%s\'`' % ( str(year), month, day )
    
    
def generatePlugin( EVT_FILE_NAME ):
    
    evtOutName = 'shm_' + EVENTID + '.evt'
    # write_event.py.tpl
    # ------------------
    fIn = open(WRITE_EVT_PY_TPL, 'r')
    fContent = fIn.read()
    fContent = fContent.replace('#evtPath#', SHSCRATCH_PATH + evtOutName)

    outFileName = fIn.name.replace('.tpl', '')
    fOut = open(outFileName, 'w')
    fOut.write(fContent)
    os.chmod(outFileName, stat.S_IRWXU)

    fOut.close()
    fIn.close()

    # EXTERNAL_PROCESS_TPL
    #---------------------
    if not os.path.isfile( EXTERNAL_PROCESS_TPL) and os.path.isfile(EXTERNAL_PROCESS):
        shutil.copy(EXTERNAL_PROCESS, EXTERNAL_PROCESS_TPL)
    shutil.copy( EXTERNAL_PROCESS_TPL, EXTERNAL_PROCESS )

        
    fIn = open(EXTERNAL_PROCESS, 'r')
    fContent = fIn.read()

    # NOTE: Absolute path must be given for the evtInFile
    # --------------------------------------------------
    fContent = fContent.replace('#evtInFile#', EVT_FILE_NAME)
    fContent = fContent.replace('#evtOutFile#', evtOutName)
    outFileName = fIn.name
    fIn.close()
    
    fOut = open(outFileName, 'w')
    fOut.write(fContent)
    fOut.close()
    

    # READ_STREAMS_TPL
    #-----------------
    fIn = open(READ_STREAMS_TPL, 'r')
    fContent = fIn.read()
    
    fContent = fContent.replace('#onsetTime#', addTimeOffset( ONSETTIMES[0] ) )
    fContent = fContent.replace('#timeWindow#', str( calcTimeWindow() ) )

    STATIONCODES.reverse() 
    stations = ','.join(STATIONCODES)
    
    fContent = fContent.replace('#stations#', stations )
    fContent = fContent.replace('#component#', 'Z')
    fContent = fContent.replace('#bandAndInstrument#', 'BH')
   
    # list stations to be hidden
    length = len(STATIONCODES)
    traceId = ','.join(map(str, range(1, length+1)))
    fContent = fContent.replace('#traceId#', traceId)

    outFileName = fIn.name.replace('.tpl', '')
    fOut = open(outFileName, 'w')
    fOut.write(fContent)

    fOut.close()
    fIn.close()

    # Build the sfdfile
    # -----------------
    wildcard = buildSfdWildcard(EVENTID)
    cmd = """
        #!/bin/bash
        
        echo "Changing directory to: $MYDATA"
        cd $MYDATA
        
        alias sfdline='%s'
        
        sfdFile=sfdfile.sfd
        
        if [ -e $sfdFile ]
        then
            rm $sfdFile
        fi    
        #ACTIVE=*`date +'%%Y.%%j'`
        ACTIVE=%s
        for i in $DATA/2006/*/{%s}/BHZ.D/$ACTIVE
        do
            echo "processing file $i"
            sfdline -seedrec=512 -byteoff=0 $i >> $sfdFile
        done
        """ % (SFDLINE, wildcard, stations)
   
    os.system(cmd)    
    
if __name__ == '__main__':
    
    # Find the latest eventfile
    # --------------------------
    evtFileName = findLatestEvtFile()
    print 'Processing event: ' + evtFileName
        
    # Read test event file
    # NOTE: The home directory of the running plugin is the home
    #       directory of seismic handler.
    # ----------------------------------------------------------
    evtFile = open(evtFileName, 'r')
    
    # Parse event file
    # ----------------
    parseEvtFile(evtFile)
    
    # Generate plugin
    # ---------------
    generatePlugin( evtFileName )
