import seis
import Numeric as Num
from seis._numeric import Biquad
import seis.numeric
import seis._numeric


class FFT_Filter(seis.numeric.FFT_Filter):

    # This is just a convenience wrapper

    def __call__(self, trace):
        self.apply(trace)

    def apply(self, trace):
        """
        Apply filter to a seis.Trace in-place
        """
        assert isinstance(trace, seis.Trace)
        seis.numeric.FFT_Filter.apply(self, trace.data, trace.fsamp)


class IIR_Filter(seis.numeric.IIR_Filter):

    # This is just a convenience wrapper

    def __call__(self, trace):
        self.apply(trace)

    def apply(self, trace):
        """
        Apply filter to a seis.Trace in-place
        """
        assert isinstance(trace, seis.Trace)
        # XXX a bit dangerous, because fsamp is not checked
        trace.data = seis.numeric.IIR_Filter.apply(self, trace.data)


epsilon=0.00001

def pole2biquad(p, fsamp, f0=None):

    # p is an analog pole

    if f0 is None:
        f0 = abs(p)/(2*Num.pi)
    K  = Num.tan(Num.pi*f0/fsamp)
    p  = p/(2*Num.pi*f0) # scale so that pole is on unit circle
    if abs(p.imag/p.real)<epsilon: # scaled pole is at -1+0j
        X = 1./(K+1)
        a0 = 1.
        a1 = (K-1)*X
        a2 = 0.
        b0 = K*X
        b1 = b0
        b2 = 0.
    else:
        # the following includes the bilinear transform
        Q = -1./(2*p.real)
        X =  1./(K*K+K/Q+1)
#       print K
        a0 = 1.
        a1 = 2*(K*K-1)*X
        a2 = (K*K-K/Q+1)*X
        b0 = K*K*X    # *Num.sqrt(2.)/K
        b1 = b0*2
        b2 = b0
    a = Num.array([a0, a1, a2])
    b = Num.array([b0, b1, b2])
    return b, a

def zero2biquad(z, fsamp, f0=None):

    b, a = pole2biquad(z, fsamp, f0=f0)
#   print "b,a = ",b, a
    b, a = a/b[0], b/b[0] # swap and normalize  XXX XXX
#   print "b0=", b[0]
#   b, a = a, b
#   print "b,a = ",b, a
    return b, a

def pz2biquads(poles, zeros, norm, fsamp=1.):

    biquads = []
    _zeros = list(zeros[:])
    _poles = list(poles[:])

    # cancel (shorten) poles and matching zeros
    for iz,z in enumerate(_zeros):
        for ip,p in enumerate(_poles):
            if p is not None and z is not None and abs(p-z) < epsilon:
                _poles[ip] = None
                _zeros[iz] = None
                break

    for xx in [ x for x in _poles if x is not None and x.imag > -epsilon ]:
        if abs(xx)<epsilon: # pole at or near origin of s plane
            x = 0.5/fsamp
            b, a = [x, x, 0], [1., -1., 0]
            biquads.append(Biquad(a,b))
        else:
            b, a = pole2biquad(xx, fsamp)
            biquads.append(Biquad(a,b))

    for xx in [ x for x in _zeros if x is not None and x.imag > -epsilon ]:
        if abs(xx)<epsilon: # zero at or near origin of s plane
            x = 2*fsamp
            b, a = [x, -x, 0], [1., 1., 0]
            biquads.append(Biquad(a,b))
        else:
            b, a = zero2biquad(xx, fsamp)
            biquads.append(Biquad(a,b))

    return biquads


class IIR_FilterBiquads(object):

    def __init__(self, poles, zeros, norm, fsamp=1.):
        self.biquads = pz2biquads(poles, zeros, norm, fsamp)

    def __call__(self, arg):
        return self.apply(arg)

    def apply(self, x):
        for filt in self.biquads:
            x = filt.apply(x)
        return x


def IIR_ButterLowPass(N, fc, fsamp=1.):
    """
    Returns a new IIR_Filter instance of a recursive Butterworth
    lowpass of order N.
    Specify fc as corner frequency, and fsamp as sampling frequency
    (both in Hz).
    """
    import seis._scipy_signal

    Wn = 2.*fc/fsamp
    b, a = seis._scipy_signal.butter(N, Wn, btype='lowpass')
    return IIR_Filter(a, b)

def IIR_ButterLowPass2(N, fc, fsamp=1.):
    import seis._numeric
    return bw_low(N, fc, fsamp) # XXX


def IIR_ButterHighPass(N, fc, fsamp=1.):
    """
    Returns a new IIR_Filter instance of a recursive Butterworth
    highpass of order N.
    Specify fc as corner frequency, and fsamp as sampling frequency
    (both in Hz).
    """
    import seis._scipy_signal

    Wn = 2.*fc/fsamp
    b, a = seis._scipy_signal.butter(N, Wn, btype='highpass')
    return IIR_Filter(a, b)

def IIR_ButterBandPass(N, f1, f2, fsamp=1.):
    """
    Returns a new IIR_Filter instance of a recursive Butterworth
    bandpass of order N.
    Specify f1 and f12 as lower and upper passband limits, and the
    sampling frequency (all in Hz).
    """
    import seis._scipy_signal

    Wn   = (2.*f1/fsamp, 2.*f2/fsamp)
    b, a = seis._scipy_signal.butter(N, Wn, btype='bandpass')
    return IIR_Filter(a, b)






def FFT_ButterLowPass(N, fc):
    # N is the number of poles
    wc   = fc*2*Num.pi
    k    = Num.arange(N)
    pole = 1j*wc*Num.exp(1j*(2*k+1)*Num.pi/(2.*N))
    zero = []
    norm = wc**N
    return FFT_Filter(pole, zero, norm)

def FFT_ButterHighPass(N, fc):
    # N is the number of poles
    wc   = fc*2*Num.pi
    k    = Num.arange(N)
    pole = 1j*wc*Num.exp(1j*(2*k+1)*Num.pi/(2.*N))
    zero = Num.array([0.+0j]*N)
    norm = 1.
    return FFT_Filter(pole, zero, norm)

def FFT_ButterBandPass(N, f1, f2):
    # N is the number of poles
    k    = Num.arange(N)
    ptmp = 1j*2*Num.pi*Num.exp(1j*(2*k+1)*Num.pi/(2.*N))
    pole = Num.concatenate([ f1*ptmp, f2*ptmp ])
    zero = Num.array([0j]*N)
    norm = (f2*2*Num.pi)**N
    return FFT_Filter(pole, zero, norm)






def FFT_WoodAnderson(input="vel"):
    """
    Creates a Wood-Anderson short-period seismometer simulation filter.
    """
    pole = [-6.283185-4.712389j, -6.283185+4.712389j,]
    zero = [ 0.+0j]
    if input[:3].lower()=="dis":
        zero.append(0.+0j)
    norm = 0.0028
#   norm = 1.
    return FFT_Filter(pole, zero, norm)

def FFT_WWSSN_SP(input="vel"):
    """
    Creates a WWSSN short-period seismometer simulation filter.
    """
    pole = [-4.009271-4.442279j, -4.009271+4.442279j,
            -4.607669-6.996659j, -4.607669+6.996659j]
    zero = [ 0.+0j ]*2
    if input[:3].lower()=="dis":
        zero.append(0.+0j)
#   norm = 70.18386
#   norm = 1
    norm = 1/0.0619919263
    return FFT_Filter(pole, zero, norm)

def FFT_WWSSN_SP2(input="vel"):
    """
    Creates a WWSSN short-period seismometer simulation filter.
    """
    pole = [-3.367788-3.731514j, -3.367788+3.731514j,
            -7.037168-4.545562j, -7.037168+4.545562j]
    zero = [ 0.+0j ]*2
    if input[:3].lower()=="dis":
        zero.append(0.+0j)
    norm = 70.18386
    return FFT_Filter(pole, zero, norm)






def hilbert(trc):
    """
    Computes the Hilbert transform of the trace in place.
    """
    assert isinstance(trc, seis.Trace)

    trc.data = seis.numeric.hilbert(trc.data)

def envelope(trc):
    """
    Computes the envelope of the trace in place.
    """
    assert isinstance(trc, seis.Trace)

    trc.data = seis.numeric.envelope(trc.data)

def instphas(trc):
    """
    Compute the instantaneous phase of the trace in place.
    """
    # UNTESTED
    assert isinstance(trc, seis.Trace)

    raw = trc.data[:]
    hilbert(trc)
    hil = trc.data
    pha = Num.arctan2(raw,hil)
    trc.data = pha

def instfreq(trc):
    """
    Compute the instantaneous frequency of the trace in place.
    """
    # UNTESTED
    assert isinstance(trc, seis.Trace)
    instphas(trc)
    # XXX

def mean(trc):

    assert isinstance(trc, seis.Trace)
    return Num.average(trc.data)


def demean(trc):

    assert isinstance(trc, seis.Trace)

    trc.data = seis.numeric.demean(trc.data)


def detrend(trc, n=0):
    """
    Determines the least squares best fit line and removes it.
    """
    assert isinstance(trc, seis.Trace)

    trc.data = seis.numeric.detrend(trc.data, n=n)


def taper(trc, t1, t2=None):

##  assert isinstance(trc, seis.Trace)
    if not t2: t2 = t1 # XXX need to check taperwidths
##    ilen = trc.fsamp/trc.nsamp
##    from seis._numeric import _taper_op
##    op = _taper_op(trc.nsamp, t1*ilen, t2*ilen)
##    tc = trc.data.typecode()
##    if tc in "fFdD":
##            trc.data *= op
##    else:   trc.data = (trc.data*op).astype(tc)

    import seis.numeric
    tc = trc.data.typecode()
    if tc in "fFdD":
        trc.data = seis.numeric.taper(trc.data, int(trc.fsamp*t1), int(trc.fsamp*t2))
        trc.data = trc.data.astype(tc)
    else:
        raise TypeError, "seis.filters.taper"


def resamp(trc, fsamp):

    assert isinstance(trc, seis.Trace)

    # XXX _cext XXX
    from seis._cext import _resamp
    trc.data, trc.fsamp = _resamp(trc.data, trc.fsamp/fsamp), fsamp


def __bwhp(n, df, order):
    # n: number of frequencies, df: normalized frequency increment
    ff = (df*Num.arange(0.,n))**(2*order)
    return ff/(1.+ff)


def __bwlp(n, df, order):
    # n: number of frequencies, df: normalized frequency increment
    ff = (df*Num.arange(0.,n))**(2*order)
    return 1./(1.+ff)


def butterworthLowPass(trc, f0, order=2):
    """
    Zero-phase Butterworth low-pass
    """
    assert isinstance(trc, seis.Trace)

    from FFT import real_fft, inverse_real_fft
    from seis.numeric import nextPowerOf2
    n, nfft = trc.nsamp, nextPowerOf2(trc.nsamp)
    df = trc.fsamp/nfft
    fx = __bwlp(nfft/2+1, df/f0, order)
    ff = real_fft(trc.data, nfft)
    ff = inverse_real_fft(ff*fx, nfft)
    trc.data = ff[:n]


def butterworthHighPass(trc, f0, order=2):
    """
    Zero-phase Butterworth high-pass
    """
    assert isinstance(trc, seis.Trace)
    from FFT import real_fft, inverse_real_fft
    from seis.numeric import nextPowerOf2
    n, nfft = trc.nsamp, nextPowerOf2(trc.nsamp)
    df = trc.fsamp/nfft
    fx = __bwhp(nfft/2+1, df/f0, order)
    ff = real_fft(trc.data, nfft)
    ff = inverse_real_fft(ff*fx, nfft)
    trc.data = ff[:n]

def gaussianLowPass(trc, a):
    """
    Apply a zero-phase Gaussian low-pass.
    """
    assert isinstance(trc, seis.Trace)

    trc.data = seis.numeric.gauss(trc.data)

def semblance(traces, nw=20):
    """
    Compute the semblance of the input traces. A single
    new trace containing the semblance trace is returned.
    """
    # XXX _cext XXX
    from seis._cext  import _semblance
    from seis.types import Trace
    f = [ t.data for t in traces ]
    Trace(traces[0])
    fs, ch = traces[0].fsamp, traces[0].cha
    s = _semblance(f, nw)
    return Trace(data=s, fsamp=fs, cha=ch)

def correctGain(traces, gain=1.E09):
    for trc in traces:
        trc.setGain(gain)
