import sys
from seis.tablefile import TabFile
from Numeric import add, arange, array, concatenate, sqrt, zeros
import seis._model

def poisson(_vpvs):
    """Computes the corresponding Poisson's ratio from a vpvs ratio."""
    return (0.5*_vpvs**2-1)/(_vpvs**2-1)

def vpvs(pr):
    """Computes the corresponding vpvs ratio from a Poisson's ratio."""
    return sqrt((1-pr)/(0.5-pr))

def _interp_z(z, n): # interpolates the layers by inserting new boundaries
    zz = []
    for i in xrange(0,len(z)-1):
        if n[i]==1: zz.append(z[i])
        if n[i]>=2: zz.extend(arange(n[i])*(z[i+1]-z[i])/n[i] + z[i])
    zz.append(z[-1])
    return array(zz)

def _interp_f(f, n): # interpolates the layer properties
    ff = []
    for i in xrange(0,len(f)-1):
        if n[i]==1: ff.append(f[i])
        if n[i]>=2: ff.extend((arange(n[i])+.5)*(f[i+1]-f[i])/n[i] + f[i])
    ff.append(f[-1])
    return array(ff)

class ModelFile1D(TabFile):

    types = {    'z':float, 'vp':float, 'vs':float, 'rh':float,
                'pr':float, 'qp':float, 'qs':float,  'n':int, 'xy':float }

    def __init__(self, f):
        assert type(f) in [str, file]
	TabFile.__init__(self, f)

class EarthModel1D(dict):

    def __init__(self, *args, **kwargs):
        """
        Can be initialized from any of the following:
            - One single argument:
                file name or file open file descriptor
                or dictionary containing the columns
            - Several keyword arguments, one keyword per column
        """

        self.isFlat = 0 # rename to 'flattened'

        columns = {}

        if args and len(args)==1:
            obj = args[0]
            # Expect obj to be either a file name or a file descriptor
            # that can be passed to the ModelFile1D constructor.
            if isinstance(obj, str):
                obj = file(obj)
            if isinstance(obj, file):
                obj = ModelFile1D(obj)
            if isinstance(obj, dict):
                for col in obj:
                    if col in ModelFile1D.types:
                        columns[col] = obj[col]
                    else:
                        raise TypeError,  "unexpected field '%s'" % col
            else:
                raise TypeError, "cannot initialize '%s' from '%s'" % \
                        (type(self), type(obj))

        for col in kwargs:
            if col in ModelFile1D.types:
                columns[col] = kwargs[col]
            else:
                raise TypeError, "__init__() got an unexpected " \
                                 "keyword argument '%s'" % col

        if 'n' in columns:
            columns['z'] = array(_interp_z(columns['z'], columns['n']))
            for col in columns:
                if not col in ['z','n']:
                    columns[col] = array(_interp_f(columns[col], columns['n']))

        for col in columns:
            self[col] = array(columns[col])


    def __getattr__(self, attrname):
            try:
                return self[attrname]
            except KeyError:
                if attrname == "d":
                    return concatenate([self['z'][1:]-self['z'][:-1], [0.]])
                if attrname == "vp":
                    vs, pr = self['vs'], self['pr']
                    if vs and pr:
                        self['vp'] = vs*vpvs(pr)
                        return self['vp']
                if attrname == "vs":
                    vp, pr = self['vp'], self['pr']
                    if vp and pr:
                        self['vs'] = vp/vpvs(pr)
                        return self['vs']
                if attrname == "rh":
                    self['rh'] = self['vp']*0.32 + 0.77
                    return self['rh']
                if attrname == "qp":
                    return array([700.]*len(self['z']))
                if attrname == "qs":
                    return array([300.]*len(self['z']))
                raise AttributeError, attrname

    def read(self, f):
        pass

    def flatten(self):
        """ Applies an Earth-flattening transform to the model in-place """
        seis._model.flatten(self.z, self.vp, self.vs, self.rh)
        self.isFlat = 1

#   def unflatten(self):
#   def plot(self):
#   def read(self):
#   def write(self):

    def trace(self, angular_slowness=6.4):
        """ Computes the delay times and offsets for each layer boundary
        by tracing a P ray and an S ray from the receiver through the
        model. Returns tp, ts, xp, xs Numeric.array's, which contain the
        times and horizontal piercing point offsets of the respective
        phases to each interface in the model. """
        p = angular_slowness/111.195
        tp, ts, xp, xs = seis._model.trace(self.z, self.vp, self.vs, p)
        return tp, ts, xp, xs

    def trace2(self, angular_slowness=6.4):
        """ Computes the delay times and offsets for each layer boundary
        by tracing a P ray and an S ray from the receiver through the
        model. Returns tp, ts, xp, xs Numeric.array's, which contain the
        times and horizontal piercing point offsets of the respective
        phases to each interface in the model. """
        p = angular_slowness/111.195
        tp, ts, xp, xs = seis._model.trace2(self.z, self.vp, self.vs, p)
        return tp, ts, xp, xs

    def delay_times(self, angular_slowness=6.4):
        """ Computes the delay times for each layer boundary. """
        p = angular_slowness/111.195
        tp, ts = seis._model.delay_times(self.z, self.vp, self.vs, p)
        return tp, ts

    def piercing_points(self, angular_slowness=6.4):
        """ Same as trace(), but only returns the piercing point offsets. """
        tp, ts, xp, xs = self.trace(angular_slowness)
        return xp, xs

    def piercing_coords(self, lat, lon, baz, angular_slowness=6.4):
        """ Like piercing_points(), but computes the geographical
        coordinates of the piercing points for a given station
        coordinate lat, lon and back azimuth baz. """
        from seis.delta import xandaz2coord
        xp, xs = self.piercing_points(angular_slowness)
        latp, lonp = xandaz2coord(xp, baz, lat, lon)
        lats, lons = xandaz2coord(xs, baz, lat, lon)
        return latp, lonp, lats, lons

#   def mocorr(self, f, fsamp, p=6.4, p0=6.4):
#       """ Moveout-corrects f in place. """
#       mocorr(f, self.z, self.vp, self.vs, p, p0, fsamp)

    def layer(self, zz):
	""" Determines the index of the layer at depth zz. """
	z, n = self.z, len(self.z)
	if zz > z[-1]: return n-1 # XXX HACK!
	i, k = 0, n-1
	while k-i>1:
	    j = (i+k)/2
	    if z[j] <= zz:  i = j
	    else:	    k = j
	return i


def Iasp91(dz=5., zmax=1000.):
    from seis._model import iasp91_arr

    z = dz*arange(zmax/dz+0.5)
    z, vp, vs, rh = iasp91_arr(z)
    mod = EarthModel1D(z=z, vp=vp, vs=vs, rh=rh)

    return mod


def _str2int(a):      return array(map(int, a))
def _str2float(a):    return array(map(float, a))

class FlatModel:

    def __init__(self, f=None, z=None, vp=None, vs=None,
			     rho=None, qp=None, qs=None):
	if f:	self.read(f)
	if z:	self.z	 = z
	if vp:	self.vp	 = vp
	if vs:	self.vs	 = vs
	if rho:	self.rho = rho
	if qp:	self.qp	 = qp
	if qs:	self.qs	 = qs
        self.isFlat = 0
 
    def __getattr__(self, attrname):
	if attrname == "d":
            return concatenate([self.z[1:]-self.z[:-1], [0.]])
	if attrname == "rh":
	    return self.rho
	if attrname == "qp":
	    return array([225.]*len(self.z))
	if attrname == "qs":
	    return array([110.]*len(self.z))
	raise AttributeError, attrname

    def read(self, f=sys.stdin, pr=0.25, dhint=1.):
        if type(f) is str:
            f = file(f)
        if type(f) is not file:
            raise ValueError, "can't read from %s" % str(f)

	lines = [ line.strip() for line in f.readlines() ]
	lines = [ line for line in lines if line[0]!='#'  ]
	first = lines[0].lower().split()
	lines = [ line.split() for line in lines[1:] ]
	prop  = {}

        for i in range(0,len(first)):
            prop[first[i]] = [ line[i] for line in lines ]
        if "z" not in prop or "n" not in prop:
            raise ValueError, "no 'z' and/or 'n' columns found in model file"
        n = array([   int(x) for x in prop["n"] ])
        z = array([ float(x) for x in prop["z"] ])
        del prop["n"], prop["z"]
        for k in prop.keys():
            prop[k] = array([ float(x) for x in prop[k] ])
        if "pr" in prop: pr = prop["pr"]
        if "vp" in prop and "vs" not in prop:
            prop["vs"] = prop["vp"]/sqrt((1-pr)/(0.5-pr))
        if "vs" in prop and "vp" not in prop:
            prop["vp"] = prop["vs"]*sqrt((1-pr)/(0.5-pr))
	if "rho" not in prop:
            prop["rho"] = prop["vp"]*0.32 + 0.77

        for i in range(0,len(n)-1): 
            if n[i]<0:
                n[i] = int((z[i+1]-z[i])/dhint) + 1
#               n[i] = int((prop["vs"][i+1]-prop["vs"][i])/0.05) + 1

	self.z = _interp_z(z, n)
	for k in prop.keys():
	    setattr(self, k, _interp_f(prop[k], n))

    def write(self, f=sys.stdout):
# XXX UNTESTED XXX
	if type(f) is str:
            f = file(f, "w")
	if type(f) is not file:
	    raise ValueError, "cannot handle %s" % str(f)

	z,vp,vs,rh,qp,qs = self.z,self.vp,self.vs,self.rh,self.qp,self.qs
        f.write ("   z      vp     vs     rho      qp     qs  n\n");
	for i in range(0, len(self.z)):
	    f.write ("%8.3f %6.3f %6.3f %6.3f %6.1f %6.1f 1\n" \
		% (z[i], vp[i], vs[i], rh[i], qp[i], qs[i]))
 
    def fromfunction(self, func, dz=5., zmax=1000.):
# XXX UNTESTED XXX
        self.n   = int(zmax/dz+1)
        self.z   = arange(0., self.n)*dz
        self.vp  = arange(0., self.n)
        self.vs  = arange(0., self.n)
        self.rho = arange(0., self.n)
        for i in range(0, self.n):
            self.vp[i], self.vs[i], self.rho[i], k = func(self.z[i])

    def flatten_old(self):
	""" performs a earth-flattening transform """
        R = 6371.
	n = len(self.z)
        r, q = R-self.z, arange(0., n)
        d, q[:-1], q[-1] = r[:-1]-r[1:], 2*R/(r[1:]+r[:-1]), R/r[-1]
        d *= q[:-1]
        self.vp *= q
        self.vs *= q
        for i in range (0, n-1):
            self.z[i+1] = self.z[i] + d[i]
        self.isFlat = 1

    def flatten(self):
        from seis._model import flatten
        flatten(self.z, self.vp, self.vs, self.rh)
        self.isFlat = 1

    def trace(self, angular_slowness=6.4):
        """
        Traces a P ray and an S ray back from the receiver through
        the model. Returns tp, ts, xp, xs Numeric.array's, which
	contain the times and horizontal piercing point offsets of
	the respective phases to each interface in the model.
        """
        from seis._cext import _trace
        p = angular_slowness/111.195
        tp, ts, xp, xs = _trace(self.d, self.vp, self.vs, p, self.isFlat)
        return tp, ts, xp, xs

    def delay_times(self, angular_slowness=6.4):
	"""
	Same as trace(), except that it only returns the times
	but not the piercing point offsets.
	"""
        from seis._cext import _delay_times
        p = angular_slowness/111.195
        tp, ts = _delay_times(self.d, self.vp, self.vs, p)
	return tp, ts

    def piercing_points(self, angular_slowness=6.4):
	"""
	Same as trace(), except that it only returns the
	piercing point offsets but not the delay times.
	"""
        from seis._cext import _trace
        p = angular_slowness/111.195
        tp, ts, xp, xs = _trace(self.d, self.vp, self.vs, p, self.isFlat)
        return xp, xs

    def piercing_coords(self, lat, lon, baz, angular_slowness=6.4):
	"""
	Like piercing_points(), but computes the geographical
	coordinates of the piercing points for a given station
	coordinate lat, lon and back azimuth baz.
	"""
	from seis.delta import xandaz2coord
	xp, xs = self.piercing_points(angular_slowness)
	latp, lonp = xandaz2coord(xp, baz, lat, lon)
	lats, lons = xandaz2coord(xs, baz, lat, lon)
        return latp,lonp,lats,lons

    def layer(self, zz):
	""" Determines the index of the layer at depth zz """
	z, n = self.z, len(self.z)
	if zz>z[-1]: return n-1 # XXX HACK!
	i, k = 0, n-1
	while k-i>1:
	    j = (i+k)/2
	    if z[j] <= zz:  i = j
	    else:	    k = j
	return i

#   def insert_boundary(self, zz):
#       """ Inserts a layer boundary at depth zz """
#       i = self.layer(zz)

def plotModel(mod, eps="model.eps", plot_range=(3.,5.,0.,60.),
	      avg=None, err=None):
    import os
    vmin, vmax, zmin, zmax = plot_range

    if   type(mod) is FlatModel:
	models = [ mod ]
    elif type(mod) is type([]):
	models = mod
    else: raise ValueError, "cannot determine model type"

    if avg:
	if err:
	    e1, e2 = FlatModel(), FlatModel()
	    e1.z = e2.z  = avg.z
	    e1.vp, e2.vp = avg.vp-err.vp, avg.vp+err.vp
	    e1.vs, e2.vs = avg.vs-err.vs, avg.vs+err.vs
	    e1.rh, e2.rh = avg.rh-err.rh, avg.rh+err.rh
	    models = models + [e1] + [e2]
	models = models + [avg]
    else:   e1 = e2 = None

    cmd_= "psxy -R%g/%g/%g/%g -JX8/-12 -P " % (vmin, vmax, zmin, zmax)
    cmd = cmd_ + "-K >'%s'" % eps
    fd = os.popen (cmd, 'w')
    for mod in models:
	if not mod: continue
	if mod is avg:
		cmd = cmd_ + "-W12/0 -K -O >>'%s'" % eps
	elif mod is e1 or mod is e2:
		cmd = cmd_ + "-W10/255/0/0 -K -O >>'%s'" % eps
	else:	cmd = cmd_ + "-W2/160 -K -O >>'%s'" % eps
	fd = os.popen (cmd, 'w')
	z, v, nlay = mod.z, mod.vs, len(mod.z)
	for i in range(0, nlay):
	    try:
		fd.write("%f %f\n" % (v[i], z[i]))
		if i+1 == nlay:
		    fd.write("%f %f\n" % (v[i], zmax))
		    break
		fd.write("%f %f\n" % (v[i], z[i+1]))
	    except:
		break
    zz = zmax-zmin
    if	 zz <=  5:  dz = (  1,  0.2)
    elif zz <= 10:  dz = (  1,  0.5)
    elif zz <= 20:  dz = (  2,  1)
    elif zz <= 40:  dz = (  5,  1)
    elif zz <= 80:  dz = ( 10,  2)
    elif zz <=200:  dz = ( 20,  5)
    elif zz <=400:  dz = ( 50, 10)
    else:	    dz = (100, 20)
    cmd = cmd_ + "-Bf0.2a1/a%ff%f::WSne -O >>'%s'" % (dz[0],dz[1],eps)
    fd = os.popen (cmd, 'w')
    return

def avgModel(models, dz=0.2, zmax=10.):
    nz = int(zmax/dz+1.5)
    z  = arange(0,nz)*dz
    vs = zeros(nz)
    zs = z + 0.5*dz
    iz = arange(0,nz)
    mm, nmm = models, len(models)
    ivp = array([ [m.vp[m.layer(zs[i])] for i in iz] for m in mm ])
    ivs = array([ [m.vs[m.layer(zs[i])] for i in iz] for m in mm ])
    irh = array([ [m.rh[m.layer(zs[i])] for i in iz] for m in mm ])

    # average model & standard deviations
    vp  = sum(ivp)/nmm
    vs  = sum(ivs)/nmm
    rh  = sum(irh)/nmm
    avg = FlatModel(z=z, vp=vp, vs=vs, rho=vs)

    vp  = sqrt(sum((ivp-vp)**2)/nmm)
    vs  = sqrt(sum((ivs-vs)**2)/nmm)
    rh  = sqrt(sum((irh-rh)**2)/nmm)
    std = FlatModel(z=z, vp=vp, vs=vs, rho=vs)

    return avg, std

def mergeModels(m1, m2, zm):
    i1, i2 = m1.layer(zm) , m2.layer(zm)
    zn	= concatenate([m2.z[:i2],  m1.z[i1:]])
    vpn = concatenate([m2.vp[:i2], m1.vp[i1:]])
    vsn = concatenate([m2.vs[:i2], m1.vs[i1:]])
    rhn = concatenate([m2.rh[:i2], m1.rh[i1:]])
    qpn = concatenate([m2.qp[:i2], m1.qp[i1:]])
    qsn = concatenate([m2.qs[:i2], m1.qs[i1:]])
    mn = FlatModel(z=zn, vp=vpn, vs=vsn, rho=rhn, qp=qpn, qs=qsn)
    return mn 
