#!/usr/bin/env python

__all__ = [ "polarization" ]
__doc__ = """
Routines for the computation of wavefield polarization

In this module techniques are used which are described in

Kanasewich, E. R. (1981). Time Sequence Analysis in Geophysics. The University
of Alberta Press. 

Kanasewich, E. R. (1990). Seismic Noise Attenuation, Band 7 von Handbook of
Geophysical Exploration. Pergamon Press, Oxford.

Vidale, J. E. (1986). Complex polarization analysis of particle motion.
BSSA, 76, 1393-1405.

Bokelmann, G. H. R. (1995). P-wave array polarization analysis and
effective anisotropy of the brittle crust, GJI, 120, 145-162.
"""

import Numeric as Num, LinearAlgebra as LA
from seis._numeric import polarization as polarization2

def __sort2(x, y, i, k):

    if x[i] < x[k]:
        x[i], x[k], y[i], y[k] = x[k], x[i], y[k], y[i]

def __sort_eigenvectors3(eigval, eigvec):

    __sort2(eigval, eigvec, 1, 2)
    __sort2(eigval, eigvec, 0, 1)
    __sort2(eigval, eigvec, 1, 2)

def polar2(fx,fy):
    """
    Computes the polarization properties of the functions fx, fy
    -> phi, lin
    """
    from seis.numeric import covar2 as covariances

    cxx, cyy, cxy = covariances(fx,fy)
    cxx, cyy, cxy = cxx.real, cyy.real, cxy.real

    p = -(cxx+cyy)
    q = cxx*cyy-cxy*cxy

    xx = Num.sqrt(0.25*p*p-q)
    x1 = -0.5*p + xx
    x2 = -0.5*p - xx

    if x1>x2:
        val1 = x1
        val2 = x2
    else:
        val1 = x2
        val2 = x1

    x2 = (val1-cxx)/cxy
    x1 = 1./Num.sqrt(1.+x2*x2)
    x2 = x1*x2

#   vec= (x1,x2)
    lin= (val1-val2)/(val1+val2)
    phi= Num.arctan2(x2, x1)*180/Num.pi

    return phi, lin 


def polarization(fx,fy,fz, rltype=1):
    """
    Computes the polarization properties of the functions fx, fy, fz
    -> inc,azi,lin

    fx, fy and fz must of course be time aligned, gain corrected and
    have the same type.

    rltype selects the method to compute the rectilinearity.
    XXX More details XXX

    inc is the incidence angle, i.e. the angle which the largest
        eigenvector forms with the the (x,y) plane
    azi is the azimuth of the largest eigenvector, i.e. the angle
        of its projection on the (x,y) plane, counted from the
        direction of fx (0 deg) over the direction of fy (90 deg)
    lin is the rectilinearity
    """
    from seis.numeric import covar3 as covariances
    covmat = covariances(fx, fy, fz)
    covmat = covmat.real # after Bataille & Chiu (1991)
    eigval, eigvec = LA.eigenvectors(covmat)
    eigval = eigval.real
    __sort_eigenvectors3(eigval, eigvec)

    if eigvec[0][2]<0:
        eigvec = -eigvec

    # rectilinearity

    if rltype in [3, 4]:
        # compute ellipticities
        eyx = Num.sqrt(eigval[1]/eigval[0]) # main ellipticity
        ezx = Num.sqrt(eigval[2]/eigval[0])
        ezy = Num.sqrt(eigval[2]/eigval[1]) # transverse ellipticity

    if rltype == 1:
        # after Kanasewich (1981, 1990)
        q = 0.5 # oder z.B. 0.5
        x = eigval[1]/eigval[0]
        if q != 1.:
            try:
                x = pow(x, q)
            except ValueError:
                import sys
                sys.stderr.write("XX %f %f\n" % (x, q))
        lin = 1. - x;

    elif rltype == 2:
        # after Vidale (1986)
        lin = 1. - (eigval[1]+eigval[2])/eigval[0]

    elif rltype == 2.1:
        # Bokelmann (1995)
        lin = 1. - 0.5*(eigval[1]+eigval[2])/eigval[0]

    elif rltype == 3:
        x = ((1-eyx)**2+(1-ezx)**2+(eyx-ezx)**2)/(2*(1+eyx+ezx)**2)
        lin = Num.sqrt(x)

    elif rltype == 4:
        lin = 1 - 1.5*(eyx+ezx)/(1+eyx+ezx)

    elif rltype == 5:
        lin = (eigval[0]-eigval[1])/(eigval[0]+eigval[1]+eigval[2])

    # "incidence" angle
    inc = Num.arctan2(Num.sqrt(eigvec[0][0]**2+eigvec[0][1]**2),
            eigvec[0][2])*180/Num.pi

    # azimuth
    azi = Num.arctan2(eigvec[0][0], eigvec[0][1])*180/Num.pi
    if inc>90.:
        inc = 180. - inc
        azi = 180. + azi
    azi = Num.fmod(azi+180, 360)

    return inc, azi, lin
