import sys, os, time, seis, seiscomp.detector, seiscomp.fe
from   seiscomp._base_db \
    import _MyDB, Config, get_config, paramstyle, xparamstyle
from seiscomp.autoloc_misc      import sorted_by_time
from   Numeric import average

table_alert_region_rect = "alert_region_rect"
table_alert_region_circ = "alert_region_circ"
table_alert_criteria    = "alert_criteria"
table_alert_users       = "alert_users"
table_alert_sent        = "alert_sent"

xcols = {
    table_alert_region_rect:
        "name, latmin, latmax, lonmin, lonmax",
    table_alert_region_circ:
        "name, latcen, loncen, radius",
    table_alert_criteria:
        "email, region_name, magmin, delay, format",
    table_alert_users:
        "email, short, realname, institution, scientific, priority",
    table_alert_sent:
        "destination, event_id"
}   

schema = {

table_alert_region_rect:
"""
CREATE TABLE alert_region_rect (
    name            VARCHAR(40)     NOT NULL,
    latmin          DECIMAL(4,1)    NOT NULL, -- bottom latitude
    latmax          DECIMAL(4,1)    NOT NULL, -- top latitude
    lonmin          DECIMAL(5,1)    NOT NULL, -- left longitude
    lonmax          DECIMAL(5,1)    NOT NULL, -- right longitude
    PRIMARY KEY (name)
);
""",

table_alert_region_circ:
"""
CREATE TABLE alert_region_circ (
-- circular regions not yet implemented
    name            VARCHAR(40)     NOT NULL,
    latcen          DECIMAL(4,1)    NOT NULL,
    loncen          DECIMAL(5,1)    NOT NULL,
    radius          DECIMAL(5,1)    NOT NULL,
    PRIMARY KEY (name)
);

""",

table_alert_criteria:
"""
CREATE TABLE alert_criteria (
-- criteria for email/sms alerts
    region_name     VARCHAR(40)     NOT NULL,
    email           CHAR(60)        NOT NULL,
    -- minimum magnitudes for email and pager alerts
    magmin_email    DECIMAL(4,2),               -- min. magn. for email alert
    magmin_pager    DECIMAL(4,2),               -- min. magn. for pager alert
    PRIMARY KEY (region_name, email)
);
""",

table_alert_users:
"""
CREATE TABLE alert_users (
-- Recipients for email/sms alerts
    email           CHAR(60)        NOT NULL,
    pager           VARCHAR(40),
    realname        VARCHAR(60),
    institution     VARCHAR(60),                -- e.g. "GFZ Potsdam"
    scientific      TINYINT DEFAULT 1,          -- scientific or not
    priority        TINYINT DEFAULT 0,          -- highest are served first
    wants_email     TINYINT DEFAULT 1,          -- email alert requested? 
    wants_pager     TINYINT DEFAULT 0,          -- pager alert requested? 
    delay           TINYINT DEFAULT 15,         -- delay after origin time
    PRIMARY KEY (email)
);
""",

table_alert_sent:
"""
CREATE TABLE alert_sent (
-- log of successfully sent alerts incl. time stamp
    destination     CHAR(60)        NOT NULL,   -- email address/phone number
    event_id        CHAR(20)        NOT NULL,
    time            TIMESTAMP,
    PRIMARY KEY (destination, event_id)
);
"""

}

xparam = {}
for x in xcols:
    xparam[x] = ",".join(len(xcols[x].split(","))*[xparamstyle])

agency = "GFZ"

class RegionRect(object):

    def __init__(self, name, latmin, latmax, lonmin, lonmax):
        self.name = name
        self.latmin, self.latmax, self.lonmin, self.lonmax = \
                        latmin, latmax, lonmin, lonmax

class RegionCirc(object):
    pass

class User(object):

    def __init__(self, email, pager=None, realname=None, institution=None,
                              scientific=True, priority=0, delay=0,
                              wants_email=True, wants_pager=False):
        self.email = email
        self.pager = pager
        self.realname = realname
        self.institution = institution
        self.scientific = scientific
        self.priority = priority
        self.delay = delay
        self.wants_email = wants_email
        self.wants_pager = wants_pager

class Criteria(object):

    def __init__(self,  region_name, email,
                        magmin_email=0., magmin_pager=None):
        self.region_name  = region_name
        self.email        = email
        self.magmin_pager = magmin_pager
        self.magmin_email = magmin_email


def format_short(origin):

    origin.load_magnitudes()
    nmag, mag = origin.summary_magnitude()
    nph = len([a for a in origin.get_associations() if a.used])
    reg = seiscomp.fe.region(origin.latitude, origin.longitude)
    ns, ew = "NS"[origin.latitude<0], "EW"[origin.longitude<0]
    dat = origin.time.strftime("%e%b%y").strip().upper()
    tim = origin.time.strftime("%T")
    lat = "%.1f%c" % (abs(origin.latitude), ns)
    lon = "%.1f%c" % (abs(origin.longitude), ew)
    dep = "%dkm"    % max(int(origin.depth), 10)

    msg = "M=%.1f\n%s\n%s %s\n%s %s %s\n%d/%d" \
        % (mag,reg,tim,dat,lat,lon,dep,nmag,nph)

    return msg


def format_long(origin):

    msg =  "Event-ID:    %s\n\n" % origin.id

    str1 = origin.time.strftime("%d %B %Y   (%j)")
    str2 = seiscomp.fe.region(origin.latitude, origin.longitude)
    msg += "%-36s  %-40s\n\n" % (str1, str2)

    _time  = str(origin.time).split()[1][:-2]
    ns, ew = "NS"[origin.latitude<0], "EW"[origin.longitude<0]
    _lat   = "%6.2f %c" % (abs(origin.latitude), ns)
    _lon   = "%6.2f %c" % (abs(origin.longitude), ew)
    _dep   = "%5.1f km" % origin.depth
    if origin.time_error is None:
        msg += "Location estimate\n\n"
        msg += " Origin time       %s\n" % _time
        msg += " Latitude          %s\n" % _lat
        msg += " Longitude         %s\n" % _lon
        msg += " Depth             %s\n" % _dep
    else:
        _time_err = "+/- %5.1f sec" % origin.time_error
        _lat_err  = "+/- %5.1f km"  % origin.latitude_error
        _lon_err  = "+/- %5.1f km"  % origin.longitude_error
        _dep_err  = "+/- %5.1f km"  % origin.depth_error
        msg += "Location estimate (+/- standard deviation)\n"
        msg += " Origin time       %10s %10s\n" % (_time, _time_err)
        msg += " Latitude          %10s %10s\n" % (_lat,  _lat_err)
        msg += " Longitude         %10s %10s\n" % (_lon,  _lon_err)
        msg += " Depth             %10s %10s\n" % (_dep,  _dep_err)
        # if loc.dep_err > 0.01:
        #     dep_err = "+/- %5.1f km" % loc.dep_err
        # else:   dep_err = "(fixed)"


#   msg += "\n Magnitudes:\n"
#   msg += "  ML=%.1f (%d)  mb=%.1f (%d)\n" \
#       % (loc.ml_avg, loc.ml_nst, loc.mb_avg, loc.mb_nst)
#   msg += "  Median=%.1f (%d)\n" % (loc.m_med, loc.m_nst)
#   msg += "\n"
#   msg += " 90-Percent confidence region ellipse:\n" \
#           "  Semi-major axis   %6.1f km  strike: %.1f deg\n" \
#           "  Semi-minor axis   %6.1f km\n" \

    magnitudes = origin.db.get_sta_magnitudes(origin.id)
    mag_types = []
    for m in magnitudes:
        if m.mag_type not in mag_types:
            mag_types.append(m.mag_type)

    mag = [ m.mag_value for m in magnitudes ]
    msg += " Magnitude           %3.1f (%d measurements)\n\n" \
        % (average(mag), len(mag))

    for mag_type in mag_types:
        mag = [ m.mag_value for m in magnitudes if m.mag_type == mag_type ]
        msg += " %-2s = %.1f  (%d)\n" % (mag_type, average(mag), len(mag))

    msg += "\n\nPhase Arrivals\n\n"
    msg += " net sta    dist azi    phase  arrival     ttime   res\n"
    assos = origin.get_associations()
    asso_pick = [(asso, origin.db.pick(asso.pick_id)) for asso in assos]
    asso_pick = [(pick.time, asso, pick) for asso,pick in asso_pick]
    asso_pick.sort()
    asso_pick = [(asso, pick) for t,asso,pick in asso_pick]
    for asso, pick in asso_pick:
        if asso.used:
            tim = str(pick.time)[11:-2]
            ttm = pick.time - origin.time
            pha = asso.phase
            sta = pick.sta
            net = pick.net
            if asso.residual is not None:
                    res = "%5.1f" % asso.residual
            else:   res = ""
            result = origin.db.get_sta_coord(net, sta)
            if result:
                station = result[0]
                d,a,b = seis.delta.delazi(origin.latitude, origin.longitude,
                                          station.lat, station.lon)
                msg += " %-3s %-5s %5.1f %3.0f    %-6s %s %6.1f %5s\n" \
                    % (net, sta, d, a, pha, tim, ttm, res)
            else: sys.stderr.write("Station %s_%s not found - skipping\n" % (net,sta))

    msg += "\n\nMagnitude Measurements\n\n"
    msg += " net sta   mtyp mval\n"
    known_mag_types = ["ML", "mb", "mB", "mCc", "Ms"]
    for mag_type in known_mag_types:
        for mag in [ m for m in magnitudes if m.mag_type == mag_type ]:
            sta = mag.sta
            net = mag.net
            typ = mag.mag_type
            val = mag.mag_value
            msg += " %-3s %-5s %-4s %4.1f\n" \
                     % (net, sta, typ, val)

    return msg

__autoloc_body = \
""" Autoloc alert %(origin_id)s: determined by %(nsta)s stations
    
 LocSAT solution (with start solution, %(nsta)s stations used, weight %(nsta)s):

  %(reg)s  %(mnam)s=%(mavg).1f  %(tim)s  %(lat).2f %(ns)c  %(lon).2f %(ew)c  %(dep).0f km

  Stat  Net   Date       Time           Amp    Per   Res  Dist   Az   mb   ML

%(list)s
"""

__autoloc_footer = \
"""An email alert for this event is sent only once. To examine later
and more accurate estimates for the same event, please look at
http://www.gfz-potsdam.de/geofon/alerts/%(origin_id)s.

You find all recent GEOFON alerts at http://geofon.gfz-potsdam.de/db/eqinfo.php

This service is a joint effort of all GEVN (GEOFON Extended Virtual Network)
partner networks (http://www.gfz-potsdam.de/geofon/new/netabs/eq_inf_nets.html).
Please give proper credit!
"""

def format_autoloc1(origin, footer=__autoloc_footer):

    origin.load_magnitudes()

    assos = origin.get_associations()
    asso_pick = [(asso, origin.db.pick(asso.pick_id)) for asso in assos]
    asso_pick = [(pick.time, asso, pick) for asso,pick in asso_pick]
    asso_pick.sort()
    asso_pick = [(asso, pick) for t,asso,pick in asso_pick if asso.used]

    d = {
        "reg":seiscomp.fe.region(origin.latitude, origin.longitude),
        "origin_id":origin.id, "nsta":len(asso_pick),
        "lat":abs(origin.latitude), "lon":abs(origin.longitude),
        "dep":origin.depth, "tim":"  ".join(str(origin.time)[:-4].split()),
        "ew":"EW"[origin.longitude<0], "ns":"NS"[origin.latitude<0] 
    }

    n_ml, ml = origin.average_magnitude("ML")
    n_mb, mb = origin.average_magnitude("mb")
    if n_mb>n_ml:
            d["mnam"], d["mavg"] = "mb", mb
    else:   d["mnam"], d["mavg"] = "ML", ml

    d["list"] = ""

    for asso, pick in asso_pick:
        if asso.used:
            tx = str(pick.time)
            sta = pick.sta
            net = pick.net
            amb, pmb = pick.ampl_mb, pick.peri_mb
            mb = ml = 0.
            for m in origin.magnitudes:
                if (m.net, m.sta) == (net, sta):
                    if   m.mag_type == "mb": mb = m.mag_value
                    elif m.mag_type == "ML": ml = m.mag_value
#           if asso.residual:
            res = "%5.1f" % asso.residual
#           else:   res = ""
            result = origin.db.get_sta_coord(net, sta)
            if not result:
#               self.warning("Station %s_%s not found - skipping\n" % (net,sta))
                continue
            station = result[0]
            dist,azi,baz = seis.delta.delazi(origin.latitude, origin.longitude,
                                             station.lat, station.lon)
            d["list"] += "  %-5s %2s  %s  %s     %8.1f  %3.1f %5s %5.1f %5.1f %4.1f %4.1f\n" % \
                    (sta, net, tx[2:10], tx[11:-2], amb, pmb, res, dist, azi, mb, ml)

    return (__autoloc_body + footer) % d

class AlertDB(_MyDB):

    def setup_tables(self, table=None, drop=False):
        if isinstance(table, str):
            tables = [ table ]
        elif table is None:
            tables = schema.keys()
        else:
            raise TypeError, "wrong table name"
 
        for table in tables:
            self.verbose("working on table '%s'\n" % table)
            if table not in schema:
                raise ValueError, "table '%s' not in schema" % table
            cur = self.cursor()
            if drop:
                sql = "DROP TABLE %s" % table
                cur.execute(sql)
            sql = schema[table]
            cur.execute(sql)

    def get_matching_regions(self, lat, lon):
        table = table_alert_region_rect
        sql   = "SELECT %s\nFROM %s\n" % (xcols[table], table)

        sql  += "WHERE (%.3f BETWEEN latmin AND latmax)\n" % lat
        sql  += " AND ((lonmin<lonmax " \
                    "AND  (%.3f BETWEEN lonmin AND lonmax))" % lon
        sql  +=   " OR (lonmin>lonmax " \
                    "AND ((%.3f BETWEEN lonmin AND 180) OR " \
                         "(%.3f BETWEEN -180 AND lonmax))))" % (lon, lon)

        result = []
        cur = self.cursor()
        cur.execute(sql)
        for name, latmin, latmax, lonmin, lonmax in cur.fetchall():
            msg = "%s %g %g %g %g\n" % (name, latmin, latmax, lonmin, lonmax)
            self.debug(msg)
#           reg = RegionRect(name, latmin, latmax, lonmin, lonmax)
            result.append(name)

        return result

    def get_recipients_by_region(self, region_name, magnitude=None):
        table = table_alert_criteria
        sql   = "SELECT email\nFROM %s\n" % table
        sql  += "WHERE region_name=\"%s\"" % region_name
        if magnitude is not None:
             sql += " AND magmin <=%.2f" % magnitude

        result = []
        cur = self.cursor()
        cur.execute(sql)
        for email, in cur.fetchall():
            self.debug("found matching email address %s\n" % email)
            result.append(email)

        return result

    def get_sent_messages(self, twin=None):
        # return event id, email/sms, type, time for alerts sent within time
        # wondow
        pass

    def add_region(self, reg):
        sql = """REPLACE INTO alert_region_rect
             (latmin,latmax,lonmin,lonmax,name)
             VALUES(%s,%s,%s,%s,%s)"""
        val = (reg.latmin, reg.latmax, reg.lonmin, reg.lonmax, reg.name)
        cur = self.cursor()
        cur.execute(sql, val)

    def add_user(self, user):
        sql = """REPLACE INTO alert_users
                 (email,pager,wants_email,wants_pager,realname,institution,scientific,priority)
                 VALUES(%s,%s,%s,%s,%s,%s,%s,%s)"""
        val = ( user.email, user.pager, user.wants_email, user.wants_pager,
                user.realname, user.institution, user.scientific, user.priority)
        cur = self.cursor()
        cur.execute(sql, val)

    def add_criteria(self, crit):
        sql = """REPLACE INTO alert_criteria
                 (region_name, email, magmin_email, magmin_pager)
                 VALUES(%s,%s,%s,%s)"""
        val = ( crit.region_name, crit.email,
                crit.magmin_email, crit.magmin_pager)
        cur = self.cursor()
        cur.execute(sql, val)

    def send_message(self, origin, recipient):
        pass

    def get_recipients_by_origin(self, origin):
        # This is the most important method
        regions = self.get_matching_regions(origin.latitude, origin.longitude)
        result = []
        for region in regions:
            magnitude = None
#           magnitude = origin.summary_magnitude # XXX idea for future XXX
            recipients = self.get_recipients_by_region(region, magnitude)
            result.extend(recipients)

        return result

    def previously_sent(self, origin, recipient=None):
        table = table_alert_sent
        sql   = "SELECT destination, time FROM %s " % table
        sql  += "WHERE event_id=\"%s\" " % origin.id # XXX must become event.id
        if recipient is not None:
             sql += "AND STRCMP(destination, \"%s\") = 0" % recipient
        result = []
        cur = self.cursor()
        cur.execute(sql)
        for destination, time in cur.fetchall():
            result.append(destination)
        return result

    def log_success(self, origin, recipient):
        table = table_alert_sent
        sql   = "REPLACE INTO %s (%s)" % (table, xcols[table])
        sql  += "VALUES (%s)" % xparam[table]

        val   = (recipient, origin.id)
        cur = self.cursor()
        cur.execute(sql, val)


    def __send_email(self, rcpt, subj, body):
        
#       import smtplib
#       smtp = smtplib.SMTP("smtp-server")
#       smtp.login("username", "topsecet")
#       msid = "20060226.233455.XhorP@geofon.gfz-potsdam.de"

#       smtp.sendmail('saul@gfz-potsdam.de (GEOFON DC)', rcpt, "To: GEOFON subscribers;\nSubject: Alert TEST\nMessage-ID: <%s>\n" % msid + msg)
        pipe = os.popen("nail -s '%s' '%s'" % (subj, rcpt), 'w')
        pipe.write("%s\n" % body)
        pipe.close()

    def __send_sms(self, number, body):
        # FIXME to be implemented XXX
        pass

    def send_alert(self, origin, rcpt=None, force=False, test=False,
                         dump=False):
        """ Send an alert email. By default, the recipients are retrieved
            from the database. They can, however, be specified explicitly.
            Then they don't have to be in the database at all. But even if
            they are not in the database, the email is logged in the database.
            if force is True, then the email is sent anyway, no matter if it
            has been sent to that recipient previously.
        """

        self.debug("Working on event %s\n" % origin.id)

        # XXX
        now = seis.time()
        wait = -(now - origin.time)/60. - 15
        if wait > 0:
            self.verbose("Event %s too new - waiting for %.0f more minutes\n" % (origin.id, wait))
            return
        # XXX

        origin.load_magnitudes()
        nmag, mag = origin.summary_magnitude()
        if nmag < 10:
            self.verbose("Event %s has too few magnitudes (%d)\n" % (origin.id, nmag))

        if rcpt is None:
            rcpt = self.get_recipients_by_origin(origin)
#           for email in rcpt:
#               if not email in self.user:
#                   self.user[rcpt] = self.get_user(rcpt)

        if isinstance(rcpt, str):
            rcpt = [ rcpt ]
        assert isinstance(rcpt, list)

#       msg  = format_long(origin)
#       msg  = format_short(origin)
        msg  = format_autoloc1(origin)
        if dump:
            print msg

        subj = "*ALERT* M%3.1f %s" % (mag, origin.region)

        # query the database once for previously sent alerts for this origin
        previously_sent = [ r.lower() for r in self.previously_sent(origin) ]
        for r in rcpt:
            rlow = r.lower()
            if not force and rlow in previously_sent:
                self.verbose("%s previously sent to %s - skipping\n" \
                        % (origin.id, r))
                continue

            if test: # test run. only report what *would* be sent
                self.write("%s would be sent to %s\n" % (origin.id, r))

            else:    # do try to send the messages 
                try:
                    self.__send_email(r, subj, msg)
                    self.log_success(origin, r)
                    previously_sent.append(rlow) # dont always re-read db
                    self.verbose("%s successfully sent to %s\n" % (origin.id,r))
                except:
                    self.error("Failed to send %s to %s\n" % (origin.id, r))
                                            

format_sms = format_short
